# mypy: disable-error-code="misc, override, var-annotated"
"""Response serializers for the various AWS protocol specifications.

There are some similarities among the different protocols with respect
to response serialization, so the code is structured in a way to avoid
code duplication where possible.  The diagram below illustrates the
inheritance hierarchy of the response serializer classes.

                           +--------------------+
                           | ResponseSerializer |
                           +--------------------+
                             ^       ^       ^
         +-------------------+       |       +---------------------+
         |                           |                             |
    +----+--------------+  +---------+----------+  +---------------+----+
    | BaseXMLSerializer |  | BaseRestSerializer |  | BaseJSONSerializer |
    +-------------------+  +--------------------+  +--------------------+
         ^             ^          ^           ^           ^           ^
         |             |          |           |           |           |
         |         +---+----------+----+  +---+-----------+----+      |
         |         | RestXMLSerializer |  | RestJSONSerializer |      |
         |         +-------------------+  +--------------------+      |
         |                                                            |
    +----+------------+                                  +------------+---+
    | QuerySerializer |                                  | JSONSerializer |
    +-----------------+                                  +----------------+
         ^
         |
    +----+----------+
    | EC2Serializer |
    +---------------+

Return Value
============

The response serializers expose a single public method: ``serialize()``
This method takes in any result (dict, object, Exception, etc.) and
returns a serialized ResponseDict of the following form:

    {
        "body": <RESPONSE_BODY>,
        "headers": <RESPONSE_HEADERS>,
        "status_code": <RESPONSE_HTTP_STATUS_CODE>,
    }

The body serialization output is text (Python strings), not bytes, mostly
for ease of inspection while debugging (pretty printing is also supported).
The exception is blob types, which are assumed to be binary and will be
encoded as ``utf-8``.

"""

from __future__ import annotations

import abc
import base64
import calendar
import json
from collections import namedtuple
from collections.abc import Callable, Generator, Mapping, MutableMapping
from dataclasses import dataclass
from datetime import datetime
from typing import (
    Any,
    Optional,
    TypedDict,
    Union,
)

import xmltodict
from botocore import xform_name
from botocore.compat import formatdate
from botocore.utils import is_json_value_header, parse_to_aware_datetime

from moto.core.errors import ErrorShape, get_error_model
from moto.core.model import (
    ListShape,
    MapShape,
    OperationModel,
    ServiceModel,
    Shape,
    StructureShape,
)
from moto.core.utils import MISSING, get_value

Serialized = MutableMapping[str, Any]


class ResponseDict(TypedDict):
    body: str
    headers: MutableMapping[str, str]
    status_code: int


class SerializationContext:
    def __init__(self, request_id: Optional[str] = None) -> None:
        self.request_id = request_id or "request-id"


class TimestampSerializer:
    TIMESTAMP_FORMAT_ISO8601 = "iso8601"
    TIMESTAMP_FORMAT_ISO8601_ZEROED = "iso8601_zeroed"
    TIMESTAMP_FORMAT_RFC822 = "rfc822"
    TIMESTAMP_FORMAT_UNIX = "unixtimestamp"

    ISO8601 = "%Y-%m-%dT%H:%M:%SZ"
    ISO8601_MICRO = "%Y-%m-%dT%H:%M:%S.%fZ"
    ISO8601_MICRO_ZEROED = "%Y-%m-%dT%H:%M:%S.000Z"

    def __init__(self, default_format: str) -> None:
        self.default_format = default_format

    def serialize(
        self, serialized: Serialized, value: Any, shape: Shape, key: str
    ) -> None:
        timestamp_format = shape.serialization.get(
            "timestampFormat", self.default_format
        )
        serialized_value = self._convert_timestamp_to_str(value, timestamp_format)
        serialized[key] = serialized_value

    def _timestamp_iso8601(self, value: datetime) -> str:
        if value.microsecond > 0:
            timestamp_format = self.ISO8601_MICRO
        else:
            timestamp_format = self.ISO8601
        return value.strftime(timestamp_format)

    def _timestamp_iso8601_zeroed(self, value: datetime) -> str:
        return value.strftime(self.ISO8601_MICRO_ZEROED)

    @staticmethod
    def _timestamp_unixtimestamp(value: datetime) -> float:
        return int(calendar.timegm(value.timetuple()))

    def _timestamp_rfc822(self, value: Union[datetime, float]) -> str:
        if isinstance(value, datetime):
            value = self._timestamp_unixtimestamp(value)
        return formatdate(value, usegmt=True)

    def _convert_timestamp_to_str(
        self, value: Union[int, str, datetime], timestamp_format: str
    ) -> str:
        timestamp_format = timestamp_format.lower()
        converter = getattr(self, f"_timestamp_{timestamp_format}")
        datetime_obj = parse_to_aware_datetime(value)  # type: ignore
        final_value = converter(datetime_obj)
        return final_value


class HeaderSerializer:
    # https://smithy.io/2.0/spec/http-bindings.html#httpheader-serialization-rules
    DEFAULT_ENCODING = "utf-8"
    DEFAULT_TIMESTAMP_FORMAT = TimestampSerializer.TIMESTAMP_FORMAT_RFC822

    def __init__(self, **kwargs: Mapping[str, Any]) -> None:
        super().__init__(**kwargs)
        self._timestamp_serializer = TimestampSerializer(self.DEFAULT_TIMESTAMP_FORMAT)

    def serialize(
        self, serialized: Serialized, value: Any, shape: Shape, key: str
    ) -> None:
        method = getattr(
            self, f"_serialize_type_{shape.type_name}", self._default_serialize
        )
        method(serialized, value, shape, key)

    @staticmethod
    def _default_serialize(
        serialized: Serialized, value: Any, _: Shape, key: str
    ) -> None:
        serialized[key] = str(value)

    def _serialize_type_boolean(
        self, serialized: Serialized, value: Any, shape: Shape, key: str
    ) -> None:
        boolean_value = "true" if value else "false"
        self._default_serialize(serialized, boolean_value, shape, key)

    def _serialize_type_list(
        self, serialized: Serialized, value: Any, shape: ListShape, key: str
    ) -> None:
        list_value = ",".join(value)
        self._default_serialize(serialized, list_value, shape, key)

    def _serialize_type_map(
        self, serialized: Serialized, value: Any, shape: MapShape, _: str
    ) -> None:
        header_prefix = shape.serialization.get("name", "")
        for key, val in value.items():
            full_key = header_prefix + key
            self._default_serialize(serialized, val, shape, full_key)

    def _serialize_type_string(
        self, serialized: Serialized, value: Any, shape: Shape, key: str
    ) -> None:
        string_value = value
        if is_json_value_header(shape):
            json_value = json.dumps(value, separators=(",", ":"))
            string_value = self._base64(json_value)
        self._default_serialize(serialized, string_value, shape, key)

    def _serialize_type_timestamp(
        self, serialized: Serialized, value: Any, shape: Shape, key: str
    ) -> None:
        wrapper = {}
        self._timestamp_serializer.serialize(wrapper, value, shape, "timestamp")
        self._default_serialize(serialized, wrapper["timestamp"], shape, key)

    def _base64(self, value: Union[str, bytes]) -> str:
        if isinstance(value, str):
            value = value.encode(self.DEFAULT_ENCODING)
        return base64.b64encode(value).strip().decode(self.DEFAULT_ENCODING)


class ResponseSerializer:
    CONTENT_TYPE = "text"
    DEFAULT_ENCODING = "utf-8"
    DEFAULT_RESPONSE_CODE = 200
    DEFAULT_ERROR_RESPONSE_CODE = 400
    # From the spec, the default timestamp format if not specified is iso8601.
    DEFAULT_TIMESTAMP_FORMAT = TimestampSerializer.TIMESTAMP_FORMAT_ISO8601
    # Clients can change this to a different MutableMapping (i.e. OrderedDict) if they want.
    # This is used in the compliance test to match the hash ordering used in the tests.
    # NOTE: This is no longer necessary because dicts post 3.6 are ordered:
    # https://stackoverflow.com/questions/39980323/are-dictionaries-ordered-in-python-3-6
    MAP_TYPE = dict

    def __init__(
        self,
        operation_model: OperationModel,
        context: Optional[SerializationContext] = None,
        pretty_print: Optional[bool] = False,
        value_picker: Any = None,
    ) -> None:
        self.operation_model = operation_model
        self.service_model = operation_model.service_model
        self.context = context or SerializationContext()
        self.pretty_print = pretty_print
        if value_picker is None:
            value_picker = DefaultAttributePicker()
        self._value_picker = value_picker
        self._timestamp_serializer = TimestampSerializer(self.DEFAULT_TIMESTAMP_FORMAT)
        self.operation_name = str(operation_model.name)
        self.path = [self.operation_name]

    def _create_default_response(self) -> ResponseDict:
        response_dict: ResponseDict = {
            "body": "",
            "headers": {},
            "status_code": self.operation_model.http.get(
                "responseCode", self.DEFAULT_RESPONSE_CODE
            ),
        }
        return response_dict

    def serialize(self, result: Any) -> ResponseDict:
        resp = self._create_default_response()
        if self._is_error_result(result):
            resp = self._serialize_error(resp, result)
        else:
            resp = self._serialize_result(resp, result)
        return resp

    def _serialize_error(
        self,
        resp: ResponseDict,
        error: Exception,
    ) -> ResponseDict:
        error_shape = get_error_model(error, self.service_model)
        serialized_error = self.MAP_TYPE()
        self._serialize_error_metadata(serialized_error, error, error_shape)
        return self._serialized_error_to_response(
            resp, error, error_shape, serialized_error
        )

    def _serialize_result(self, resp: ResponseDict, result: Any) -> ResponseDict:
        output_shape = self.operation_model.output_shape
        serialized_result = self.MAP_TYPE()
        if output_shape is not None:
            assert isinstance(output_shape, StructureShape)  # mypy hint
            self._serialize(serialized_result, result, output_shape, "")
        return self._serialized_result_to_response(
            resp, result, output_shape, serialized_result
        )

    def _serialized_error_to_response(
        self,
        resp: ResponseDict,
        error: Exception,
        shape: ErrorShape,
        serialized_error: MutableMapping[str, Any],
    ) -> ResponseDict:
        raise NotImplementedError("Must be implemented in subclass.")

    def _serialized_result_to_response(
        self,
        resp: ResponseDict,
        result: Any,
        shape: Optional[StructureShape],
        serialized_result: MutableMapping[str, Any],
    ) -> ResponseDict:
        raise NotImplementedError("Must be implemented in subclass.")

    def _serialize_error_metadata(
        self,
        serialized: Serialized,
        error: Exception,
        shape: ErrorShape,
    ) -> None:
        raise NotImplementedError("Must be implemented in subclass.")

    def _serialize_body(self, body: Any) -> str:
        raise NotImplementedError("Must be implemented in subclass.")

    # Some extra utility methods subclasses can use.
    @staticmethod
    def _is_error_result(result: object) -> bool:
        return isinstance(result, Exception)

    def _base64(self, value: Union[str, bytes]) -> str:
        if isinstance(value, str):
            value = value.encode(self.DEFAULT_ENCODING)
        return base64.b64encode(value).strip().decode(self.DEFAULT_ENCODING)

    def get_value(self, value: Any, key: str, shape: Shape) -> Any:
        context = AttributePickerContext(
            obj=value,
            key=key,
            shape=shape,
            operation_model=self.operation_model,
            service_model=self.operation_model.service_model,
            key_path=".".join(self.path),
        )
        return self._value_picker(context)

    #
    # Default serializers for the various model Shape types.
    # These can be overridden in subclasses to provide protocol-specific implementations.
    #
    def _serialize(
        self, serialized: Serialized, value: Any, shape: Shape, key: str
    ) -> None:
        method = getattr(
            self, f"_serialize_type_{shape.type_name}", self._default_serialize
        )
        if not key:
            self.path.append(shape.name)
        else:
            self.path.append(key)
        method(serialized, value, shape, key)
        self.path.pop()

    def _default_serialize(
        self, serialized: Serialized, value: Any, shape: Shape, key: str
    ) -> None:
        serialization_key = self.get_serialized_name(shape, key)
        serialized[serialization_key] = value

    def _serialize_type_structure(
        self, serialized: Serialized, value: Any, shape: StructureShape, key: str
    ) -> None:
        if value is None:
            return
        if key:
            wrapper: Any = self.MAP_TYPE()
        else:
            wrapper = serialized
        for member_key, member_shape in shape.members.items():
            member_shape.parent = shape  # type: ignore[attr-defined]
            self._serialize_structure_member(wrapper, value, member_shape, member_key)
        if key:
            self._default_serialize(serialized, wrapper, shape, key)

    def _serialize_structure_member(
        self, serialized: Serialized, value: Any, shape: Shape, key: str
    ) -> None:
        member_value = self.get_value(value, key, shape)
        if member_value is not None:
            self._serialize(serialized, member_value, shape, key)

    def _serialize_type_map(
        self, serialized: Serialized, value: Any, shape: MapShape, key: str
    ) -> None:
        key_shape = shape.key
        assert isinstance(key_shape, Shape)
        value_shape = shape.value
        assert isinstance(value_shape, Shape)
        map_list = []
        for k, v in value.items():
            wrapper = {"__current__": {}}
            self._serialize(wrapper["__current__"], k, key_shape, "key")
            self._serialize(wrapper["__current__"], v, value_shape, "value")
            map_list.append(wrapper["__current__"])
        if shape.is_flattened:
            self._default_serialize(serialized, map_list, shape, key)
        else:
            self._default_serialize(serialized, {"entry": map_list}, shape, key)

    def _serialize_type_timestamp(
        self, serialized: Serialized, value: Any, shape: Shape, key: str
    ) -> None:
        value_wrapper = {}
        value_key = "timestamp"
        self._timestamp_serializer.serialize(value_wrapper, value, shape, value_key)
        self._default_serialize(serialized, value_wrapper[value_key], shape, key)

    def _serialize_type_blob(
        self, serialized: Serialized, value: Any, shape: Shape, key: str
    ) -> None:
        blob_value = self._base64(value)
        self._default_serialize(serialized, blob_value, shape, key)

    def _serialize_type_integer(
        self, serialized: Serialized, value: Any, shape: Shape, key: str
    ) -> None:
        integer_value = int(value)
        self._default_serialize(serialized, integer_value, shape, key)

    _serialize_type_long = _serialize_type_integer

    def _serialize_type_float(
        self, serialized: Serialized, value: Any, shape: Shape, key: str
    ) -> None:
        integer_value = float(value)
        self._default_serialize(serialized, integer_value, shape, key)

    _serialize_type_double = _serialize_type_float

    def get_serialized_name(self, shape: Shape, default_name: str) -> str:
        return shape.serialization.get("name", default_name)


class BaseJSONSerializer(ResponseSerializer):
    APPLICATION_AMZ_JSON = "application/x-amz-json-{version}"
    DEFAULT_TIMESTAMP_FORMAT = "unixtimestamp"

    def _serialized_result_to_response(
        self,
        resp: ResponseDict,
        result: Any,
        shape: Optional[StructureShape],
        serialized_result: MutableMapping[str, Any],
    ) -> ResponseDict:
        resp["body"] = self._serialize_body(serialized_result)
        resp["headers"]["Content-Type"] = self._get_protocol_specific_content_type()
        return resp

    def _serialized_error_to_response(
        self,
        resp: ResponseDict,
        error: Exception,
        shape: ErrorShape,
        serialized_error: MutableMapping[str, Any],
    ) -> ResponseDict:
        resp["body"] = self._serialize_body(serialized_error)
        status_code = shape.metadata.get("error", {}).get(
            "httpStatusCode", self.DEFAULT_ERROR_RESPONSE_CODE
        )
        resp["status_code"] = status_code
        error_code = self._get_protocol_specific_error_code(shape)
        resp["headers"]["X-Amzn-Errortype"] = error_code
        resp["headers"]["Content-Type"] = self._get_protocol_specific_content_type()
        self._serialize_query_compatible_error_to_response(resp, shape)
        return resp

    def _serialize_query_compatible_error_to_response(
        self, resp: ResponseDict, shape: ErrorShape
    ) -> None:
        if "awsQueryCompatible" not in self.service_model.metadata:
            return
        fault = "Sender" if shape.is_sender_fault else "Receiver"
        resp["headers"]["x-amzn-query-error"] = f"{shape.error_code};{fault}"

    def _get_protocol_specific_content_type(self) -> str:
        content_type = self.CONTENT_TYPE
        service_model = self.operation_model.service_model
        protocol = service_model.protocol
        if protocol == "json":
            json_version = service_model.metadata.get("jsonVersion", "1.0")
            content_type = self.APPLICATION_AMZ_JSON.format(version=json_version)
        return content_type

    def _get_protocol_specific_error_code(
        self,
        error: ErrorShape,
    ) -> str:
        # https://smithy.io/2.0/aws/protocols/aws-json-1_1-protocol.html#operation-error-serialization
        service_metadata = self.operation_model.service_model.metadata
        json_version = service_metadata.get("jsonVersion")
        error_code = error.name
        prefix = (
            error.namespace
            or service_metadata.get("errorNamespace")
            or service_metadata.get("targetPrefix")
        )
        if json_version == "1.0" and prefix is not None:
            error_code = prefix + "#" + error_code
        return error_code

    def _serialize_error_metadata(
        self,
        serialized: Serialized,
        error: Exception,
        shape: ErrorShape,
    ) -> None:
        error_code = self._get_protocol_specific_error_code(shape)
        serialized["__type"] = error_code
        message = getattr(error, "message", None) or str(error)
        if shape is not None:
            self._serialize(serialized, error, shape, "")
        if message:
            serialized["Message"] = message

    def _serialize_body(self, body: Mapping[str, Any]) -> str:
        body_encoded = json.dumps(body, indent=4 if self.pretty_print else None)
        return body_encoded

    def _serialize_type_map(
        self, serialized: Serialized, value: Any, shape: MapShape, key: str
    ) -> None:
        map_obj = self.MAP_TYPE()
        self._default_serialize(serialized, map_obj, shape, key)
        for sub_key, sub_value in value.items():
            assert isinstance(shape.value, Shape)  # mypy hint
            self._serialize(map_obj, sub_value, shape.value, sub_key)

    def _serialize_type_list(
        self, serialized: Serialized, value: Any, shape: ListShape, key: str
    ) -> None:
        list_obj = []
        for list_item in value:
            wrapper = {}
            # The JSON list serialization is the only case where we aren't
            # setting a key on a dict.  We handle this by using
            # a __current__ key on a wrapper dict to serialize each
            # list item before appending it to the serialized list.
            assert isinstance(shape.member, Shape)  # mypy hint
            item_key = shape.member.name
            self._serialize(wrapper, list_item, shape.member, item_key)
            if item_key in wrapper:
                list_obj.append(wrapper[item_key])
            else:
                list_obj.append(list_item)
        self._default_serialize(serialized, list_obj, shape, key)

    def _serialize_type_structure(
        self, serialized: Serialized, value: Any, shape: StructureShape, key: str
    ) -> None:
        if shape.is_document_type:
            serialized[key] = value
            return
        super()._serialize_type_structure(serialized, value, shape, key)


class BaseXMLSerializer(ResponseSerializer):
    CONTENT_TYPE = "text/xml"

    def _serialize_namespace_attribute(self, serialized: Serialized) -> None:
        if (
            self.CONTENT_TYPE == "text/xml"
            and "xmlNamespace" in self.operation_model.metadata
        ):
            namespace = self.operation_model.metadata["xmlNamespace"]
            serialized["@xmlns"] = namespace

    def _serialized_error_to_response(
        self,
        resp: ResponseDict,
        error: Exception,
        shape: ErrorShape,
        serialized_error: MutableMapping[str, Any],
    ) -> ResponseDict:
        error_wrapper = {
            "ErrorResponse": {
                "Error": serialized_error,
                "RequestId": self.context.request_id,
            }
        }
        self._serialize_namespace_attribute(error_wrapper["ErrorResponse"])
        resp["body"] = self._serialize_body(error_wrapper)
        status_code = shape.metadata.get("error", {}).get(
            "httpStatusCode", self.DEFAULT_ERROR_RESPONSE_CODE
        )
        resp["status_code"] = status_code
        resp["headers"]["Content-Type"] = self.CONTENT_TYPE
        return resp

    def _serialized_result_to_response(
        self,
        resp: ResponseDict,
        result: Any,
        shape: Optional[StructureShape],
        serialized_result: MutableMapping[str, Any],
    ) -> ResponseDict:
        if shape and "payload" in shape.serialization:
            result_key = shape.serialization["payload"]
        else:
            result_key = f"{self.operation_model.name}Response"
        result_wrapper = {
            result_key: serialized_result,
        }
        self._serialize_namespace_attribute(result_wrapper[result_key])
        resp["body"] = self._serialize_body(result_wrapper)
        resp["headers"]["Content-Type"] = self.CONTENT_TYPE
        return resp

    def _serialize_error_metadata(
        self,
        serialized: Serialized,
        error: Exception,
        shape: ErrorShape,
    ) -> None:
        serialized["Type"] = "Sender" if shape.is_sender_fault else "Receiver"
        serialized["Code"] = shape.error_code
        message = getattr(error, "message", None)
        if shape.query_compatible_error_message:
            message = shape.query_compatible_error_message
        if message is not None:
            serialized["Message"] = message
        # Serialize any error model attributes.
        self._serialize(serialized, error, shape, "")

    def _serialize_body(self, body: Serialized) -> str:
        body_encoded = xmltodict.unparse(
            body,
            full_document=False,
            short_empty_elements=True,
            pretty=self.pretty_print,
        )
        return body_encoded

    #
    # https://smithy.io/2.0/aws/protocols/aws-query-protocol.html#xml-shape-serialization
    #
    def _serialize_type_boolean(
        self, serialized: Serialized, value: Any, shape: Shape, key: str
    ) -> None:
        # We're slightly more permissive here than we should be because the
        # moto backends are not consistent in how they store boolean values.
        # TODO: This should eventually be turned into a strict `is True` check.
        boolean_conditions = [
            (value is True),
            (str(value).lower() == "true"),
        ]
        boolean_value = "true" if any(boolean_conditions) else "false"
        self._default_serialize(serialized, boolean_value, shape, key)

    def _serialize_type_integer(
        self, serialized: Serialized, value: Any, shape: Shape, key: str
    ) -> None:
        integer_value = int(value)
        self._default_serialize(serialized, integer_value, shape, key)

    def _serialize_type_list(
        self, serialized: Serialized, value: Any, shape: ListShape, key: str
    ) -> None:
        assert isinstance(shape.member, Shape)  # mypy hinting
        list_obj = []
        for list_item in value:
            wrapper = {}
            self._serialize(wrapper, list_item, shape.member, shape.member.name)
            if wrapper:
                item_key = self.get_serialized_name(shape.member, shape.member.name)
                value = wrapper[item_key]
                if value != {}:
                    list_obj.append(value)
        if not list_obj:  # empty list serialized as "" in XML
            self._default_serialize(serialized, "", shape, key)
            return
        if shape.is_flattened:
            self._default_serialize(serialized, list_obj, shape.member, key)
        else:
            items_name = self.get_serialized_name(shape.member, "member")
            self._default_serialize(serialized, {items_name: list_obj}, shape, key)

    _serialize_type_long = _serialize_type_integer

    def _serialize_type_string(
        self, serialized: Serialized, value: Any, shape: Shape, key: str
    ) -> None:
        string_value = str(value)
        self._default_serialize(serialized, string_value, shape, key)


class BaseRestSerializer(ResponseSerializer):
    EMPTY_BODY: Serialized = ResponseSerializer.MAP_TYPE()
    REQUIRES_EMPTY_BODY = False

    def _serialized_result_to_response(
        self,
        resp: ResponseDict,
        result: Any,
        shape: Optional[StructureShape],
        serialized_result: MutableMapping[str, Any],
    ) -> ResponseDict:
        if "payload" in serialized_result:
            # Payload trumps all and is delivered as-is.
            resp["body"] = serialized_result["payload"]
        else:
            if not serialized_result["body"]:
                if self.REQUIRES_EMPTY_BODY:
                    resp["body"] = self._serialize_body(self.EMPTY_BODY)
            else:
                resp = super()._serialized_result_to_response(
                    resp, result, shape, serialized_result.get("body", {})
                )
        if "headers" in serialized_result:
            resp["headers"].update(serialized_result["headers"])
        resp["headers"]["Content-Type"] = self.CONTENT_TYPE
        return resp

    def _serialize_result(self, resp: ResponseDict, result: Any) -> ResponseDict:
        output_shape = self.operation_model.output_shape
        serialized_result = {
            "body": {},
            "headers": {},
        }
        if output_shape is not None:
            assert isinstance(output_shape, StructureShape)
            self._serialize(serialized_result, result, output_shape, "")
            payload_member = output_shape.serialization.get("payload")
            if payload_member is not None:
                payload_shape = output_shape.members[payload_member]
                payload_value = self.get_value(result, payload_member, payload_shape)
                self._serialize_payload(serialized_result, payload_value, payload_shape)

        return self._serialized_result_to_response(
            resp, result, output_shape, serialized_result
        )

    def _serialize_payload(
        self,
        serialized: Serialized,
        payload: Any,
        payload_shape: Shape,
    ) -> None:
        if payload_shape.type_name in ["blob", "string"]:
            # If it's streaming, then the body is just the value of the payload.
            serialized["payload"] = payload
        else:
            # If there's a payload member, we serialize only that one member to the body.
            serialized["body"] = self.MAP_TYPE()
            self._serialize(serialized["body"], payload, payload_shape, "")

    def _serialize_structure_member(
        self, serialized: Serialized, value: Any, shape: Shape, key: str
    ) -> None:
        if shape.is_not_bound_to_body:
            if shape.is_http_header_trait and "headers" in serialized:
                member_value = self.get_value(value, key, shape)
                if member_value is not None:
                    key_name = self.get_serialized_name(shape, key)
                    header_serializer = HeaderSerializer()
                    header_serializer.serialize(
                        serialized["headers"], member_value, shape, key_name
                    )
        elif "body" in serialized:
            if not serialized["body"]:
                serialized["body"] = self.MAP_TYPE()
            # we're at the top-level structure
            super()._serialize_structure_member(serialized["body"], value, shape, key)
        else:
            # we're in nested structure
            super()._serialize_structure_member(serialized, value, shape, key)


class RestXMLSerializer(BaseRestSerializer, BaseXMLSerializer):
    DEFAULT_TIMESTAMP_FORMAT = TimestampSerializer.TIMESTAMP_FORMAT_ISO8601

    def _serialize_body(self, body: Mapping[str, Any]) -> str:
        body_serialized = xmltodict.unparse(
            body,
            full_document=True,
            pretty=self.pretty_print,
            short_empty_elements=False,
        )
        return body_serialized


class RestJSONSerializer(BaseRestSerializer, BaseJSONSerializer):
    CONTENT_TYPE = "application/json"
    REQUIRES_EMPTY_BODY = True


class JSONSerializer(BaseJSONSerializer):
    pass


class QuerySerializer(BaseXMLSerializer):
    def _serialized_error_to_response(
        self,
        resp: ResponseDict,
        error: Exception,
        shape: ErrorShape,
        serialized_error: MutableMapping[str, Any],
    ) -> ResponseDict:
        error_wrapper = {
            "ErrorResponse": {
                "Error": serialized_error,
                "RequestId": self.context.request_id,
            }
        }
        self._serialize_namespace_attribute(error_wrapper["ErrorResponse"])
        resp["body"] = self._serialize_body(error_wrapper)
        status_code = shape.metadata.get("error", {}).get(
            "httpStatusCode", self.DEFAULT_ERROR_RESPONSE_CODE
        )
        resp["status_code"] = status_code
        resp["headers"]["Content-Type"] = self.CONTENT_TYPE
        return resp

    def _serialized_result_to_response(
        self,
        resp: ResponseDict,
        result: Any,
        shape: Optional[StructureShape],
        serialized_result: MutableMapping[str, Any],
    ) -> ResponseDict:
        response_key = f"{self.operation_model.name}Response"
        response_wrapper = {response_key: {}}
        if shape is not None:
            result_key = shape.serialization.get("resultWrapper", f"{shape.name}Result")
            response_wrapper[response_key][result_key] = serialized_result
        response_wrapper[response_key]["ResponseMetadata"] = {
            "RequestId": self.context.request_id
        }
        self._serialize_namespace_attribute(response_wrapper[response_key])
        resp["body"] = self._serialize_body(response_wrapper)
        resp["headers"]["Content-Type"] = self.CONTENT_TYPE
        return resp

    def _serialize_type_map(
        self, serialized: Serialized, value: Any, shape: MapShape, key: str
    ) -> None:
        key_shape = shape.key
        assert isinstance(key_shape, Shape)
        value_shape = shape.value
        assert isinstance(value_shape, Shape)
        map_list = []
        for k, v in value.items():
            # Query protocol does not serialize null values
            if v is None:
                continue
            wrapper = {"__current__": {}}
            self._serialize(wrapper["__current__"], k, key_shape, "key")
            self._serialize(wrapper["__current__"], v, value_shape, "value")
            map_list.append(wrapper["__current__"])
        if shape.is_flattened:
            self._default_serialize(serialized, map_list, shape, key)
        else:
            self._default_serialize(serialized, {"entry": map_list}, shape, key)


class QueryJSONSerializer(QuerySerializer):
    """Specialized case for query protocol requests that contain ContentType=JSON parameter."""

    CONTENT_TYPE = "application/json"

    def _serialized_error_to_response(
        self,
        resp: ResponseDict,
        error: Exception,
        shape: ErrorShape,
        serialized_error: MutableMapping[str, Any],
    ) -> ResponseDict:
        error_wrapper = {
            "Error": serialized_error,
            "RequestId": self.context.request_id,
        }
        resp["body"] = self._serialize_body(error_wrapper)
        status_code = shape.metadata.get("error", {}).get(
            "httpStatusCode", self.DEFAULT_ERROR_RESPONSE_CODE
        )
        resp["status_code"] = status_code
        resp["headers"]["Content-Type"] = self.CONTENT_TYPE
        return resp

    def _serialize_body(self, body: Mapping[str, Any]) -> str:
        body_encoded = json.dumps(body, indent=4 if self.pretty_print else None)
        return body_encoded

    def _serialize_type_boolean(
        self, serialized: Serialized, value: Any, shape: Shape, key: str
    ) -> None:
        # We're slightly more permissive here than we should be because the
        # moto backends are not consistent in how they store boolean values.
        # TODO: This should eventually be turned into a strict `is True` check.
        boolean_value = True if value in [True, "True", "true"] else False
        self._default_serialize(serialized, boolean_value, shape, key)

    def _serialize_type_list(
        self, serialized: Serialized, value: Any, shape: ListShape, key: str
    ) -> None:
        list_obj = []
        serialized[key] = list_obj
        for list_item in value:
            wrapper = {}
            assert isinstance(shape.member, Shape)  # mypy hint
            item_key = shape.member.name
            self._serialize(wrapper, list_item, shape.member, item_key)
            if item_key in wrapper:
                list_obj.append(wrapper[item_key])
            else:
                list_obj.append(list_item)


class EC2Serializer(QuerySerializer):
    DEFAULT_TIMESTAMP_FORMAT = TimestampSerializer.TIMESTAMP_FORMAT_ISO8601_ZEROED

    def _serialize_body(self, body: Mapping[str, Any]) -> str:
        body_serialized = xmltodict.unparse(
            body,
            full_document=True,
            pretty=self.pretty_print,
            short_empty_elements=True,
        )
        return body_serialized

    def _serialize_error_metadata(
        self,
        serialized: MutableMapping[str, Any],
        error: Exception,
        shape: ErrorShape,
    ) -> None:
        serialized["Code"] = shape.error_code
        message = getattr(error, "message", None)
        if message is not None:
            serialized["Message"] = message
        # Serialize any error model attributes.
        self._serialize(serialized, error, shape, "")

    def _serialized_error_to_response(
        self,
        resp: ResponseDict,
        error: Exception,
        shape: ErrorShape,
        serialized_error: MutableMapping[str, Any],
    ) -> ResponseDict:
        error_wrapper = {
            "Response": {
                "Errors": [{"Error": serialized_error}],
                "RequestID": self.context.request_id,
            }
        }
        self._serialize_namespace_attribute(error_wrapper["Response"])
        resp["body"] = self._serialize_body(error_wrapper)
        status_code = shape.metadata.get("error", {}).get(
            "httpStatusCode", self.DEFAULT_ERROR_RESPONSE_CODE
        )
        resp["status_code"] = status_code
        resp["headers"]["Content-Type"] = self.CONTENT_TYPE
        return resp

    def _serialized_result_to_response(
        self,
        resp: ResponseDict,
        result: Any,
        shape: StructureShape,
        serialized_result: MutableMapping[str, Any],
    ) -> ResponseDict:
        response_key = f"{self.operation_model.name}Response"
        result_wrapper = {
            response_key: serialized_result,
        }
        result_wrapper[response_key]["requestId"] = self.context.request_id
        self._serialize_namespace_attribute(result_wrapper[response_key])
        resp["body"] = self._serialize_body(result_wrapper)
        resp["headers"]["Content-Type"] = self.CONTENT_TYPE
        return resp


DoublePassEncoding = namedtuple(
    "DoublePassEncoding", ["char", "marker", "escape_sequence"]
)


class DoublePassEncoder:
    """Facilitates double pass encoding of special characters in a string
    by replacing them with markers in one pass and then replacing the markers
    with escape sequences in a second pass."""

    def __init__(self, encodings: list[DoublePassEncoding]) -> None:
        self.encodings = encodings

    def mark(self, value: str) -> str:
        for item in self.encodings:
            value = value.replace(item.char, item.marker)
        return value

    def escape(self, value: str) -> str:
        for item in self.encodings:
            value = value.replace(item.marker, item.escape_sequence)
        return value


class SqsQuerySerializer(QuerySerializer):
    """
    Special handling of SQS Query protocol responses:
        * support aws.protocols#awsQueryCompatible trait after switch to JSON protocol
        * escape HTML entities within XML tag text
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.encoder = DoublePassEncoder(
            [
                DoublePassEncoding(
                    char="\r",
                    marker="__CARRIAGE_RETURN_MARKER__",
                    escape_sequence="&#xD;",
                ),
                DoublePassEncoding(
                    char='"', marker="__DOUBLE_QUOTE_MARKER__", escape_sequence="&quot;"
                ),
            ]
        )

    def _default_serialize(
        self, serialized: Serialized, value: Any, shape: Shape, key: str
    ) -> None:
        if isinstance(value, str):
            value = self.encoder.mark(value)
        super()._default_serialize(serialized, value, shape, key)

    def _serialize_body(self, body: Serialized) -> str:
        body_encoded = super()._serialize_body(body)
        body_escaped = self.encoder.escape(body_encoded)
        return body_escaped

    def get_serialized_name(self, shape: Shape, default_name: str) -> str:
        serialized_name = super().get_serialized_name(shape, default_name)
        if self.service_model.is_query_compatible:
            serialized_name = shape.serialization.get(
                "locationNameForQueryCompatibility", serialized_name
            )
        return serialized_name


SERIALIZERS = {
    "ec2": EC2Serializer,
    "json": JSONSerializer,
    "query": QuerySerializer,
    "query-json": QueryJSONSerializer,
    "rest-json": RestJSONSerializer,
    "rest-xml": RestXMLSerializer,
}
SERVICE_SPECIFIC_SERIALIZERS = {
    "sqs": {
        "query": SqsQuerySerializer,
    }
}


def get_serializer_class(service_name: str, protocol: str) -> type[ResponseSerializer]:
    if service_name in SERVICE_SPECIFIC_SERIALIZERS:
        if protocol in SERVICE_SPECIFIC_SERIALIZERS[service_name]:
            return SERVICE_SPECIFIC_SERIALIZERS[service_name][protocol]
    return SERIALIZERS[protocol]


@dataclass
class AttributePickerContext:
    obj: Any
    key: str
    shape: Shape
    operation_model: OperationModel | None = None
    service_model: ServiceModel | None = None
    key_path: str = ""


class DefaultAttributePicker:
    def __call__(self, context: AttributePickerContext) -> Any:
        return get_value(context.obj, context.key, None)


class AttributePicker(DefaultAttributePicker):
    """Uses alias providers to find the value of an attribute in a Python object"""

    def __init__(
        self,
        alias_providers: list[type[AttributeAliasProvider]] | None = None,
        response_transformers: dict[str, Callable[[Any], Any]] | None = None,
    ) -> None:
        self.alias_providers = (
            alias_providers if alias_providers is not None else DEFAULT_ALIAS_PROVIDERS
        )
        self.response_transformers = (
            response_transformers if response_transformers is not None else {}
        )

    def __call__(self, context: AttributePickerContext) -> Any:
        obj = context.obj
        for possible_key in self.get_possible_keys(context):
            value = get_value(obj, possible_key, MISSING)
            if value is not MISSING:
                break
        else:
            value = None
        key_path = f"{context.key_path}.{context.key}"
        for transform_path, transform in self.response_transformers.items():
            if key_path.endswith(transform_path):
                value = transform(value)
                break
        return value

    def get_possible_keys(self, context: AttributePickerContext) -> Generator[str]:
        key = context.key
        for alias_provider_cls in self.alias_providers:
            alias_provider = alias_provider_cls(context)
            if alias_provider.has_alias(key):
                alias = alias_provider.get_alias(key)
                yield alias


class XFormedAttributePicker(AttributePicker):
    """Can be injected into a ResponseSerializer to aid in plucking AWS model
    attributes specified in `camelCase` or `PascalCase` from Python objects
    with standard `snake_case` attribute names.

    For a model attribute named `DBInstanceIdentifier`, this class will check
    for the following attributes on the provided object:
       * `DBInstanceIdentifier`
       * `db_instance_identifier`
    If the provided object is a class named `DBInstance`, this class will also
    check for the following attribute on the provided object:
       * `identifier`

    ``botocore.xform_name`` is used to transform the attribute name.
    """

    def get_possible_keys(self, context: AttributePickerContext) -> Generator[str]:
        for possible_key in super().get_possible_keys(context):
            yield possible_key
            yield xform_name(possible_key)


class AttributeAliasProvider(abc.ABC):
    """Abstract base class for providing attribute key aliases."""

    def __init__(self, context: AttributePickerContext) -> None:
        self.context = context

    @abc.abstractmethod
    def has_alias(self, key: str) -> bool:
        """Check if a key alias exists for the given context."""
        raise NotImplementedError()

    @abc.abstractmethod
    def get_alias(self, key: str) -> str:
        """Get the key alias for the given context."""
        raise NotImplementedError()


class ExplicitAlias(AttributeAliasProvider):
    """Provides a key alias explicitly defined in the source object's Meta class."""

    def has_alias(self, key: str) -> bool:
        obj = self.context.obj
        if not hasattr(obj, "Meta"):
            return False
        if not hasattr(obj.Meta, "serialization_aliases"):
            return False
        return key in obj.Meta.serialization_aliases

    def get_alias(self, key: str) -> str:
        return self.context.obj.Meta.serialization_aliases[key]


class NoAlias(AttributeAliasProvider):
    """Provides the key as is, without any aliasing."""

    def has_alias(self, key: str) -> bool:
        return True

    def get_alias(self, key: str) -> str:
        return key


class ModelAlias(AttributeAliasProvider):
    """Provides a key alias based on the botocore model's serialization name."""

    def has_alias(self, key: str) -> bool:
        shape = self.context.shape
        if shape is not None:
            serialization_key = shape.serialization.get("name", key)
            if serialization_key != key:
                return True
        return False

    def get_alias(self, key: str) -> Any:
        return self.context.shape.serialization["name"]


class ShapePrefixAlias(AttributeAliasProvider):
    """Provides a shortened key alias if key is prefixed with the model name.

    Example: `DBInstanceIdentifier` becomes `Identifier` if the model name is `DBInstance`.
    """

    def has_alias(self, key: str) -> bool:
        shape = self.context.shape
        if shape is not None:
            if hasattr(shape, "parent") and shape.parent.type_name == "structure":
                if key.lower().startswith(shape.parent.name.lower()):
                    # Alias is valid if it doesn't conflict with a sibling key.
                    return self.get_alias(key) not in shape.parent.members
        return False

    def get_alias(self, key: str) -> Any:
        shape = self.context.shape
        assert hasattr(shape, "parent")
        assert isinstance(shape.parent, Shape)  # mypy hint
        return key[len(shape.parent.name) :]


class ClassPrefixAlias(AttributeAliasProvider):
    """Provides a shortened key alias if key is prefixed with the source object's class name.

    Example: `DBInstanceIdentifier` becomes `Identifier` if the class name is `DBInstance`.
    """

    def has_alias(self, key: str) -> bool:
        obj = self.context.obj
        if hasattr(obj, "__class__"):
            class_name = obj.__class__.__name__
            if key.lower().startswith(class_name.lower()):
                return True
        return False

    def get_alias(self, key: str) -> Any:
        class_name = self.context.obj.__class__.__name__
        short_key = key[len(class_name) :]
        return short_key


class ShapeNameAlias(AttributeAliasProvider):
    """Provides a key alias based on the shape's name if it differs from the key."""

    def has_alias(self, key: str) -> bool:
        shape = self.context.shape
        if shape is None:
            return False
        if shape.type_name not in ["list", "structure"]:
            return False
        try:
            if shape.name in shape.parent.members:  # type: ignore[attr-defined]
                # If the name of the shape conflicts with a sibling key,
                # we don't want to use it as an alias because it will
                # pick up the sibling value.
                return False
        except AttributeError:
            pass
        if key != shape.name:
            return True
        return False

    def get_alias(self, key: str) -> Any:
        return self.context.shape.name


# Ordering is important here, as the first alias provider that matches will be used.
# We want to try the most specific alias providers first, and fall back to the more generic ones.
DEFAULT_ALIAS_PROVIDERS = [
    ExplicitAlias,
    ShapeNameAlias,
    NoAlias,
    ShapePrefixAlias,
    ClassPrefixAlias,
    ModelAlias,
]


# Response transformers can be used to modify the value of a key in the response.
def never_return(_: Any) -> None:
    """
    A utility function that is used to ensure that certain attributes are never returned
    in the response. This is useful for attributes that should not be exposed or are not
    relevant in the context of the response.
    """
    return None


def return_if_not_empty(value: Any) -> Any:
    """
    A utility function that returns the value if it is not empty (i.e., not None, "", {}, or []),
    otherwise returns None. This is useful for attributes that should only be included in the
    response if they have a meaningful value.
    """
    return value if value not in [None, "", {}, []] else None


def url_encode(value: Any) -> Any:
    """A utility function that url encodes a value before inclusion in a response."""
    from urllib.parse import quote

    return quote(value) if isinstance(value, str) else value



"""Handles incoming kinesisanalyticsv2 requests, invokes methods, returns responses."""

import json

from moto.core.responses import BaseResponse

from .models import KinesisAnalyticsV2Backend, kinesisanalyticsv2_backends


class KinesisAnalyticsV2Response(BaseResponse):
    """Handler for KinesisAnalyticsV2 requests and responses."""

    def __init__(self) -> None:
        super().__init__(service_name="kinesisanalyticsv2")

    @property
    def kinesisanalyticsv2_backend(self) -> KinesisAnalyticsV2Backend:
        """Return backend instance specific for this region."""
        return kinesisanalyticsv2_backends[self.current_account][self.region]

    def create_application(self) -> str:
        application_name = self._get_param("ApplicationName")
        application_description = self._get_param("ApplicationDescription")
        runtime_environment = self._get_param("RuntimeEnvironment")
        service_execution_role = self._get_param("ServiceExecutionRole")
        application_configuration = self._get_param("ApplicationConfiguration")
        cloud_watch_logging_options = self._get_param("CloudWatchLoggingOptions")
        tags = self._get_param("Tags")
        application_mode = self._get_param("ApplicationMode")
        application_detail = self.kinesisanalyticsv2_backend.create_application(
            application_name=application_name,
            application_description=application_description,
            runtime_environment=runtime_environment,
            service_execution_role=service_execution_role,
            application_configuration=application_configuration,
            cloud_watch_logging_options=cloud_watch_logging_options,
            tags=tags,
            application_mode=application_mode,
        )
        return json.dumps({"ApplicationDetail": application_detail})

    def list_tags_for_resource(self) -> str:
        params = json.loads(self.body)
        resource_arn = params.get("ResourceARN")
        tags = self.kinesisanalyticsv2_backend.list_tags_for_resource(
            resource_arn=resource_arn,
        )
        return json.dumps({"Tags": tags})

    def tag_resource(self) -> str:
        params = json.loads(self.body)
        resource_arn = params.get("ResourceARN")
        tags = params.get("Tags")
        self.kinesisanalyticsv2_backend.tag_resource(
            resource_arn=resource_arn, tags=tags
        )
        return json.dumps({})

    def describe_application(self) -> str:
        application_name = self._get_param("ApplicationName")
        application_detail = self.kinesisanalyticsv2_backend.describe_application(
            application_name=application_name,
        )
        return json.dumps({"ApplicationDetail": application_detail})

    def list_applications(self) -> str:
        application_summaries = self.kinesisanalyticsv2_backend.list_applications()
        return json.dumps({"ApplicationSummaries": application_summaries})



from moto.core.responses import ActionResult, EmptyResult
from moto.ec2.exceptions import InvalidParameterValueErrorUnknownAttribute
from moto.ec2.utils import add_tag_specification

from ._base_response import EC2BaseResponse


class ElasticNetworkInterfaces(EC2BaseResponse):
    def create_network_interface(self) -> ActionResult:
        subnet_id = self._get_param("SubnetId")
        private_ip_address = self._get_param("PrivateIpAddress")
        private_ip_addresses = self._get_param("PrivateIpAddresses", [])
        ipv6_addresses = self._get_param("Ipv6Addresses", [])
        ipv6_addresses = [i["Ipv6Address"] for i in ipv6_addresses]
        ipv6_address_count = self._get_int_param("Ipv6AddressCount", 0)
        secondary_ips_count = self._get_param("SecondaryPrivateIpAddressCount")
        groups = self._get_param("Groups", [])
        subnet = self.ec2_backend.get_subnet(subnet_id)
        description = self._get_param("Description")
        tags = add_tag_specification(self._get_param("TagSpecifications", []))

        self.error_on_dryrun()

        eni = self.ec2_backend.create_network_interface(
            subnet,
            private_ip_address,
            private_ip_addresses,
            groups,
            description,
            tags,
            secondary_ips_count=secondary_ips_count,
            ipv6_addresses=ipv6_addresses,
            ipv6_address_count=ipv6_address_count,
        )
        result = {"NetworkInterface": eni}
        return ActionResult(result)

    def delete_network_interface(self) -> ActionResult:
        eni_id = self._get_param("NetworkInterfaceId")

        self.error_on_dryrun()

        self.ec2_backend.delete_network_interface(eni_id)
        return EmptyResult()

    def describe_network_interface_attribute(self) -> ActionResult:
        eni_id = self._get_param("NetworkInterfaceId")
        attribute = self._get_param("Attribute")

        self.error_on_dryrun()

        eni = self.ec2_backend.get_all_network_interfaces([eni_id])[0]

        result = {"NetworkInterfaceId": eni.id}
        if attribute == "description":
            result["Description"] = {"Value": eni.description}
        elif attribute == "groupSet":
            result["Groups"] = eni.group_set
        elif attribute == "sourceDestCheck":
            result["SourceDestCheck"] = {"Value": eni.source_dest_check}
        elif attribute == "attachment":
            result["Attachment"] = eni.attachment
        else:
            raise InvalidParameterValueErrorUnknownAttribute(attribute)
        return ActionResult(result)

    def describe_network_interfaces(self) -> ActionResult:
        eni_ids = self._get_param("NetworkInterfaceIds", [])
        filters = self._filters_from_querystring()

        self.error_on_dryrun()

        enis = self.ec2_backend.get_all_network_interfaces(eni_ids, filters)
        result = {"NetworkInterfaces": enis}
        return ActionResult(result)

    def attach_network_interface(self) -> ActionResult:
        eni_id = self._get_param("NetworkInterfaceId")
        instance_id = self._get_param("InstanceId")
        device_index = self._get_param("DeviceIndex")

        self.error_on_dryrun()

        attachment = self.ec2_backend.attach_network_interface(
            eni_id, instance_id, device_index
        )
        result = {
            "AttachmentId": attachment.attachment_id,
            "NetworkCardIndex": attachment.network_card_index,
        }
        return ActionResult(result)

    def detach_network_interface(self) -> ActionResult:
        attachment_id = self._get_param("AttachmentId")

        self.error_on_dryrun()

        self.ec2_backend.detach_network_interface(attachment_id)
        return EmptyResult()

    def modify_network_interface_attribute(self) -> ActionResult:
        eni_id = self._get_param("NetworkInterfaceId")
        group_ids = self._get_param("Groups", [])
        source_dest_check = self._get_param("SourceDestCheck.Value")
        description = self._get_param("Description.Value")

        self.error_on_dryrun()

        self.ec2_backend.modify_network_interface_attribute(
            eni_id, group_ids, source_dest_check, description
        )
        return EmptyResult()

    def reset_network_interface_attribute(self) -> str:
        self.error_on_dryrun()

        raise NotImplementedError(
            "ElasticNetworkInterfaces(AmazonVPC).reset_network_interface_attribute is not yet implemented"
        )

    def assign_private_ip_addresses(self) -> ActionResult:
        eni_id = self._get_param("NetworkInterfaceId")
        secondary_ips_count = self._get_int_param("SecondaryPrivateIpAddressCount", 0)
        private_ip_addresses = self._get_param("PrivateIpAddresses", [])
        eni = self.ec2_backend.assign_private_ip_addresses(
            eni_id, private_ip_addresses, secondary_ips_count
        )
        result = {
            "NetworkInterfaceId": eni.id,
            "AssignedPrivateIpAddresses": eni.private_ip_addresses,
        }
        return ActionResult(result)

    def unassign_private_ip_addresses(self) -> ActionResult:
        eni_id = self._get_param("NetworkInterfaceId")
        private_ip_address = self._get_param("PrivateIpAddresses", [])
        self.ec2_backend.unassign_private_ip_addresses(eni_id, private_ip_address)
        return EmptyResult()

    def assign_ipv6_addresses(self) -> ActionResult:
        eni_id = self._get_param("NetworkInterfaceId")
        ipv6_count = self._get_int_param("Ipv6AddressCount", 0)
        ipv6_addresses = self._get_param("Ipv6Addresses", [])
        eni = self.ec2_backend.assign_ipv6_addresses(eni_id, ipv6_addresses, ipv6_count)
        result = {
            "AssignedIpv6Addresses": eni.ipv6_addresses,
            "NetworkInterfaceId": eni.id,
        }
        return ActionResult(result)

    def unassign_ipv6_addresses(self) -> ActionResult:
        eni_id = self._get_param("NetworkInterfaceId")
        ips = self._get_param("Ipv6Addresses", [])
        unassigned_ipv6_addresses = self.ec2_backend.unassign_ipv6_addresses(
            eni_id, ips
        )
        result = {
            "NetworkInterfaceId": eni_id,
            "UnassignedIpv6Addresses": unassigned_ipv6_addresses,
        }
        return ActionResult(result)



from moto.core.responses import ActionResult
from moto.core.utils import get_value

from ._base_response import EC2BaseResponse


class Fleets(EC2BaseResponse):
    def delete_fleets(self) -> ActionResult:
        fleet_ids = self._get_param("FleetIds", [])
        terminate_instances = self._get_bool_param("TerminateInstances")
        fleets = self.ec2_backend.delete_fleets(fleet_ids, terminate_instances)
        result = {
            "SuccessfulFleetDeletions": [
                {
                    "CurrentFleetState": fleet.state,
                    "PreviousFleetState": "active",
                    "FleetId": fleet.id,
                }
                for fleet in fleets
            ],
            "UnsuccessfulFleetDeletions": [],
        }
        return ActionResult(result)

    def describe_fleet_instances(self) -> ActionResult:
        fleet_id = self._get_param("FleetId")
        fleet_instances = self.ec2_backend.describe_fleet_instances(fleet_id)
        # The ec2_backend.describe_fleet_instances method returns a list of incompatible types,
        # objects and dictionaries, so we have to use `get_value` to pull from either.
        # TODO: This should be addressed on the backend.
        result = {
            "ActiveInstances": [
                {
                    "InstanceId": get_value(fi, "instance.id", None),
                    "InstanceType": get_value(fi, "instance.instance_type", None),
                    "SpotInstanceRequestId": get_value(fi, "id", None),
                    "InstanceHealth": "healthy",
                }
                for fi in fleet_instances
            ],
            "FleetId": fleet_id,
        }
        return ActionResult(result)

    def describe_fleets(self) -> ActionResult:
        fleet_ids = self._get_param("FleetIds", [])
        fleets = self.ec2_backend.describe_fleets(fleet_ids)
        result = {"Fleets": fleets}
        return ActionResult(result)

    def create_fleet(self) -> ActionResult:
        on_demand_options = self._get_param("OnDemandOptions", {})
        spot_options = self._get_param("SpotOptions", {})
        target_capacity_specification = self._get_param(
            "TargetCapacitySpecification", {}
        )
        launch_template_configs = self._get_param("LaunchTemplateConfigs", [])
        excess_capacity_termination_policy = self._get_param(
            "ExcessCapacityTerminationPolicy"
        )
        replace_unhealthy_instances = self._get_param("ReplaceUnhealthyInstances")
        terminate_instances_with_expiration = self._get_param(
            "TerminateInstancesWithExpiration", if_none=True
        )
        fleet_type = self._get_param("Type", if_none="maintain")
        valid_from = self._get_param("ValidFrom")
        valid_until = self._get_param("ValidUntil")

        tag_specifications = self._get_param("TagSpecifications", [])

        request = self.ec2_backend.create_fleet(
            on_demand_options=on_demand_options,
            spot_options=spot_options,
            target_capacity_specification=target_capacity_specification,
            launch_template_configs=launch_template_configs,
            excess_capacity_termination_policy=excess_capacity_termination_policy,
            replace_unhealthy_instances=replace_unhealthy_instances,
            terminate_instances_with_expiration=terminate_instances_with_expiration,
            fleet_type=fleet_type,
            valid_from=valid_from,
            valid_until=valid_until,
            tag_specifications=tag_specifications,
        )
        result = {"FleetId": request.id}
        if request.fleet_type == "instant":
            # On Demand and Spot Instances are stored as incompatible types on the backend,
            # so we have to do some extra work here.
            # TODO: This should be addressed on the backend.
            on_demand_instances = [
                {
                    "Lifecycle": "on-demand",
                    "InstanceIds": [instance["instance"].id],
                    "InstanceType": instance["instance"].instance_type,
                }
                for instance in request.on_demand_instances
            ]
            spot_requests = [
                {
                    "Lifecycle": "spot",
                    "InstanceIds": [instance.instance.id],
                    "InstanceType": instance.instance.instance_type,
                }
                for instance in request.spot_requests
            ]
            result["Instances"] = on_demand_instances + spot_requests
        return ActionResult(result)



import copy
import re
import string
from collections.abc import Iterator
from typing import Any

from moto.core.utils import iso_8601_datetime_with_milliseconds
from moto.moto_api._internal import mock_random as random


def random_id(size: int = 13) -> str:
    chars = list(range(10)) + list(string.ascii_uppercase)
    return "".join(str(random.choice(chars)) for x in range(size))


def random_cluster_id() -> str:
    return f"j-{random_id()}"


def random_step_id() -> str:
    return f"s-{random_id()}"


def random_instance_group_id() -> str:
    return f"i-{random_id()}"


class ReleaseLabel:
    version_re = re.compile(r"^emr-(\d+)\.(\d+)\.(\d+)$")

    def __init__(self, release_label: str):
        major, minor, patch = self.parse(release_label)

        self.major = major
        self.minor = minor
        self.patch = patch

    @classmethod
    def parse(cls, release_label: str) -> tuple[int, int, int]:
        if not release_label:
            raise ValueError(f"Invalid empty ReleaseLabel: {release_label}")

        match = cls.version_re.match(release_label)
        if not match:
            raise ValueError(f"Invalid ReleaseLabel: {release_label}")

        major, minor, patch = match.groups()

        return int(major), int(minor), int(patch)

    def __str__(self) -> str:
        version = f"emr-{self.major}.{self.minor}.{self.patch}"
        return version

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({str(self)})"

    def __iter__(self) -> Iterator[tuple[int, int, int]]:
        return iter((self.major, self.minor, self.patch))  # type: ignore

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.major == other.major
            and self.minor == other.minor
            and self.patch == other.patch
        )

    def __ne__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return tuple(self) != tuple(other)

    def __lt__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return tuple(self) < tuple(other)

    def __le__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return tuple(self) <= tuple(other)

    def __gt__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return tuple(self) > tuple(other)

    def __ge__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return tuple(self) >= tuple(other)


class EmrManagedSecurityGroup:
    class Kind:
        MASTER = "Master"
        SLAVE = "Slave"
        SERVICE = "Service"

    kind = ""

    group_name = ""
    short_name = ""
    desc_fmt = "{short_name} for Elastic MapReduce created on {created}"

    @classmethod
    def description(cls) -> str:
        created = iso_8601_datetime_with_milliseconds()
        return cls.desc_fmt.format(short_name=cls.short_name, created=created)


class EmrManagedMasterSecurityGroup(EmrManagedSecurityGroup):
    kind = EmrManagedSecurityGroup.Kind.MASTER
    group_name = "ElasticMapReduce-Master-Private"
    short_name = "Master"


class EmrManagedSlaveSecurityGroup(EmrManagedSecurityGroup):
    kind = EmrManagedSecurityGroup.Kind.SLAVE
    group_name = "ElasticMapReduce-Slave-Private"
    short_name = "Slave"


class EmrManagedServiceAccessSecurityGroup(EmrManagedSecurityGroup):
    kind = EmrManagedSecurityGroup.Kind.SERVICE
    group_name = "ElasticMapReduce-ServiceAccess"
    short_name = "Service access"


class EmrSecurityGroupManager:
    MANAGED_RULES_EGRESS = [
        {
            "group_name_or_id": EmrManagedSecurityGroup.Kind.MASTER,
            "from_port": None,
            "ip_protocol": "-1",
            "ip_ranges": [{"CidrIp": "0.0.0.0/0"}],
            "to_port": None,
            "source_groups": [],
        },
        {
            "group_name_or_id": EmrManagedSecurityGroup.Kind.SLAVE,
            "from_port": None,
            "ip_protocol": "-1",
            "ip_ranges": [{"CidrIp": "0.0.0.0/0"}],
            "to_port": None,
            "source_groups": [],
        },
        {
            "group_name_or_id": EmrManagedSecurityGroup.Kind.SERVICE,
            "from_port": 8443,
            "ip_protocol": "tcp",
            "ip_ranges": [],
            "to_port": 8443,
            "source_groups": [
                {"GroupId": EmrManagedSecurityGroup.Kind.MASTER},
                {"GroupId": EmrManagedSecurityGroup.Kind.SLAVE},
            ],
        },
    ]

    MANAGED_RULES_INGRESS = [
        {
            "group_name_or_id": EmrManagedSecurityGroup.Kind.MASTER,
            "from_port": 0,
            "ip_protocol": "tcp",
            "ip_ranges": [],
            "to_port": 65535,
            "source_groups": [
                {"GroupId": EmrManagedSecurityGroup.Kind.MASTER},
                {"GroupId": EmrManagedSecurityGroup.Kind.SLAVE},
            ],
        },
        {
            "group_name_or_id": EmrManagedSecurityGroup.Kind.MASTER,
            "from_port": 8443,
            "ip_protocol": "tcp",
            "ip_ranges": [],
            "to_port": 8443,
            "source_groups": [{"GroupId": EmrManagedSecurityGroup.Kind.SERVICE}],
        },
        {
            "group_name_or_id": EmrManagedSecurityGroup.Kind.MASTER,
            "from_port": 0,
            "ip_protocol": "udp",
            "ip_ranges": [],
            "to_port": 65535,
            "source_groups": [
                {"GroupId": EmrManagedSecurityGroup.Kind.MASTER},
                {"GroupId": EmrManagedSecurityGroup.Kind.SLAVE},
            ],
        },
        {
            "group_name_or_id": EmrManagedSecurityGroup.Kind.MASTER,
            "from_port": -1,
            "ip_protocol": "icmp",
            "ip_ranges": [],
            "to_port": -1,
            "source_groups": [
                {"GroupId": EmrManagedSecurityGroup.Kind.MASTER},
                {"GroupId": EmrManagedSecurityGroup.Kind.SLAVE},
            ],
        },
        {
            "group_name_or_id": EmrManagedSecurityGroup.Kind.SLAVE,
            "from_port": 0,
            "ip_protocol": "tcp",
            "ip_ranges": [],
            "to_port": 65535,
            "source_groups": [
                {"GroupId": EmrManagedSecurityGroup.Kind.MASTER},
                {"GroupId": EmrManagedSecurityGroup.Kind.SLAVE},
            ],
        },
        {
            "group_name_or_id": EmrManagedSecurityGroup.Kind.SLAVE,
            "from_port": 8443,
            "ip_protocol": "tcp",
            "ip_ranges": [],
            "to_port": 8443,
            "source_groups": [{"GroupId": EmrManagedSecurityGroup.Kind.SERVICE}],
        },
        {
            "group_name_or_id": EmrManagedSecurityGroup.Kind.SLAVE,
            "from_port": 0,
            "ip_protocol": "udp",
            "ip_ranges": [],
            "to_port": 65535,
            "source_groups": [
                {"GroupId": EmrManagedSecurityGroup.Kind.MASTER},
                {"GroupId": EmrManagedSecurityGroup.Kind.SLAVE},
            ],
        },
        {
            "group_name_or_id": EmrManagedSecurityGroup.Kind.SLAVE,
            "from_port": -1,
            "ip_protocol": "icmp",
            "ip_ranges": [],
            "to_port": -1,
            "source_groups": [
                {"GroupId": EmrManagedSecurityGroup.Kind.MASTER},
                {"GroupId": EmrManagedSecurityGroup.Kind.SLAVE},
            ],
        },
        {
            "group_name_or_id": EmrManagedSecurityGroup.Kind.SERVICE,
            "from_port": 9443,
            "ip_protocol": "tcp",
            "ip_ranges": [],
            "to_port": 9443,
            "source_groups": [{"GroupId": EmrManagedSecurityGroup.Kind.MASTER}],
        },
    ]

    def __init__(self, ec2_backend: Any, vpc_id: str):
        self.ec2 = ec2_backend
        self.vpc_id = vpc_id

    def manage_security_groups(
        self,
        master_security_group: str,
        slave_security_group: str,
        service_access_security_group: str,
    ) -> tuple[Any, Any, Any]:
        group_metadata = [
            (
                master_security_group,
                EmrManagedSecurityGroup.Kind.MASTER,
                EmrManagedMasterSecurityGroup,
            ),
            (
                slave_security_group,
                EmrManagedSecurityGroup.Kind.SLAVE,
                EmrManagedSlaveSecurityGroup,
            ),
            (
                service_access_security_group,
                EmrManagedSecurityGroup.Kind.SERVICE,
                EmrManagedServiceAccessSecurityGroup,
            ),
        ]
        managed_groups: dict[str, Any] = {}
        for name, kind, defaults in group_metadata:
            managed_groups[kind] = self._get_or_create_sg(name, defaults)
        self._add_rules_to(managed_groups)
        return (
            managed_groups[EmrManagedSecurityGroup.Kind.MASTER],
            managed_groups[EmrManagedSecurityGroup.Kind.SLAVE],
            managed_groups[EmrManagedSecurityGroup.Kind.SERVICE],
        )

    def _get_or_create_sg(self, sg_id: str, defaults: Any) -> Any:
        find_sg = self.ec2.get_security_group_by_name_or_id
        create_sg = self.ec2.create_security_group
        group_id_or_name = sg_id or defaults.group_name
        group = find_sg(group_id_or_name, self.vpc_id)
        if group is None:
            if group_id_or_name != defaults.group_name:
                raise ValueError(
                    f"The security group '{group_id_or_name}' does not exist"
                )
            group = create_sg(defaults.group_name, defaults.description(), self.vpc_id)
        return group

    def _add_rules_to(self, managed_groups: dict[str, Any]) -> None:
        rules_metadata = [
            (self.MANAGED_RULES_EGRESS, self.ec2.authorize_security_group_egress),
            (self.MANAGED_RULES_INGRESS, self.ec2.authorize_security_group_ingress),
        ]
        for rules, add_rule in rules_metadata:
            rendered_rules = self._render_rules(rules, managed_groups)
            for rule in rendered_rules:
                from moto.ec2.exceptions import InvalidPermissionDuplicateError

                try:
                    add_rule(vpc_id=self.vpc_id, **rule)
                except InvalidPermissionDuplicateError:
                    # If the rule already exists, we can just move on.
                    pass

    @staticmethod
    def _render_rules(  # type: ignore[misc]
        rules: Any, managed_groups: dict[str, Any]
    ) -> list[dict[str, Any]]:
        rendered_rules = copy.deepcopy(rules)
        for rule in rendered_rules:
            rule["group_name_or_id"] = managed_groups[rule["group_name_or_id"]].id
            rule["source_groups"] = [
                {"GroupId": managed_groups[group.get("GroupId")].id}
                for group in rule["source_groups"]
            ]
        return rendered_rules



"""codedeploy base URL and path."""

from .responses import CodeDeployResponse

url_bases = [
    r"https?://codedeploy\.(.+)\.amazonaws\.com",
]

url_paths = {
    "{0}/$": CodeDeployResponse.dispatch,
    "{0}/list-tags-for-resource$": CodeDeployResponse.list_tags_for_resource,
    "{0}/tag-resource$": CodeDeployResponse.tag_resource,
    "{0}/untag-resource$": CodeDeployResponse.untag_resource,
}



from typing import Any

from moto.core.responses import ActionResult, EmptyResult

from ._base_response import EC2BaseResponse


def parse_sg_attributes_from_dict(sg_attributes: dict[str, Any]) -> tuple[Any, ...]:
    ip_protocol = sg_attributes.get("IpProtocol")
    from_port = sg_attributes.get("FromPort")
    to_port = sg_attributes.get("ToPort")

    ip_ranges: list[dict[str, Any]] = []
    ip_ranges += sg_attributes.get("IpRanges", [])
    ip_ranges += sg_attributes.get("Ipv6Ranges", [])  # type: ignore[no-redef]

    if "CidrIp" in sg_attributes:
        cidr_ip = sg_attributes.get("CidrIp")  # type: ignore
        ip_ranges.append({"CidrIp": cidr_ip})

    if "CidrIpv6" in sg_attributes:
        cidr_ipv6 = sg_attributes.get("CidrIpv6")  # type: ignore
        ip_ranges.append({"CidrIpv6": cidr_ipv6})

    source_groups: list[dict[str, Any]] = sg_attributes.get("UserIdGroupPairs", [])
    prefix_list_ids: list[dict[str, Any]] = sg_attributes.get("PrefixListIds", [])

    return ip_protocol, from_port, to_port, ip_ranges, source_groups, prefix_list_ids


class SecurityGroups(EC2BaseResponse):
    def _process_rules_from_querystring(self) -> Any:
        group_name_or_id = self._get_param("GroupName") or self._get_param("GroupId")
        security_rule_ids = self._get_param("SecurityGroupRuleIds", [])
        sg_rule_tags = self._parse_tag_specification().get("security-group-rule", {})

        if "IpPermissions" not in self._get_params():
            # Handle single rule syntax
            (
                ip_protocol,
                from_port,
                to_port,
                ip_ranges,
                source_groups,
                prefix_list_ids,
            ) = parse_sg_attributes_from_dict(self._get_params())

            yield {
                "group_name_or_id": group_name_or_id,
                "ip_protocol": ip_protocol,
                "from_port": from_port,
                "to_port": to_port,
                "ip_ranges": ip_ranges,
                "sgrule_tags": sg_rule_tags,
                "source_groups": source_groups,
                "prefix_list_ids": prefix_list_ids,
                "security_rule_ids": security_rule_ids,
            }

        ip_permissions = self._get_param("IpPermissions", [])
        for ip_permission in ip_permissions:
            (
                ip_protocol,
                from_port,
                to_port,
                ip_ranges,
                source_groups,
                prefix_list_ids,
            ) = parse_sg_attributes_from_dict(ip_permission)

            yield {
                "group_name_or_id": group_name_or_id,
                "ip_protocol": ip_protocol,
                "from_port": from_port,
                "to_port": to_port,
                "ip_ranges": ip_ranges,
                "sgrule_tags": sg_rule_tags,
                "source_groups": source_groups,
                "prefix_list_ids": prefix_list_ids,
                "security_rule_ids": security_rule_ids,
            }

    def authorize_security_group_egress(self) -> ActionResult:
        self.error_on_dryrun()

        for kwargs in self._process_rules_from_querystring():
            rules, _ = self.ec2_backend.authorize_security_group_egress(**kwargs)
        result = {"Return": True, "SecurityGroupRules": rules}
        return ActionResult(result)

    def authorize_security_group_ingress(self) -> ActionResult:
        self.error_on_dryrun()

        for kwargs in self._process_rules_from_querystring():
            rules, _ = self.ec2_backend.authorize_security_group_ingress(**kwargs)
        result = {"Return": True, "SecurityGroupRules": rules}
        return ActionResult(result)

    def create_security_group(self) -> ActionResult:
        name = self._get_param("GroupName")
        description = self._get_param("Description")
        vpc_id = self._get_param("VpcId")
        tags = self._parse_tag_specification().get("security-group", {})

        self.error_on_dryrun()

        group = self.ec2_backend.create_security_group(
            name, description, vpc_id=vpc_id, tags=tags
        )
        result = {
            "GroupId": group.id,
            "SecurityGroupArn": group.arn,
        }
        return ActionResult(result)

    def delete_security_group(self) -> ActionResult:
        # TODO this should raise an error if there are instances in the group.
        # See
        # http://docs.aws.amazon.com/AWSEC2/latest/APIReference/ApiReference-query-DeleteSecurityGroup.html

        name = self._get_param("GroupName")
        sg_id = self._get_param("GroupId")

        self.error_on_dryrun()

        if name:
            self.ec2_backend.delete_security_group(name)
        elif sg_id:
            self.ec2_backend.delete_security_group(group_id=sg_id)

        return EmptyResult()

    def describe_security_groups(self) -> ActionResult:
        groupnames = self._get_param("GroupNames", [])
        group_ids = self._get_param("GroupIds", [])
        filters = self._filters_from_querystring()

        groups = self.ec2_backend.describe_security_groups(
            group_ids=group_ids, groupnames=groupnames, filters=filters
        )

        result = {"SecurityGroups": groups}
        return ActionResult(result)

    def describe_security_group_rules(self) -> ActionResult:
        self.error_on_dryrun()
        sg_rule_ids = self._get_param("SecurityGroupRuleIds", [])
        filters = self._filters_from_querystring()
        rules = self.ec2_backend.describe_security_group_rules(sg_rule_ids, filters)
        result = {"SecurityGroupRules": rules}
        return ActionResult(result)

    def revoke_security_group_egress(self) -> ActionResult:
        self.error_on_dryrun()

        for args in self._process_rules_from_querystring():
            # we don't need this parameter to revoke
            del args["sgrule_tags"]
            self.ec2_backend.revoke_security_group_egress(**args)
        return ActionResult({"Return": True})

    def revoke_security_group_ingress(self) -> ActionResult:
        self.error_on_dryrun()

        for args in self._process_rules_from_querystring():
            # we don't need this parameter to revoke
            del args["sgrule_tags"]
            self.ec2_backend.revoke_security_group_ingress(**args)
        return ActionResult({"Return": True})

    def modify_security_group_rules(self) -> ActionResult:
        self.error_on_dryrun()

        rules = {}
        security_group_rules_param = self._get_param("SecurityGroupRules", [])
        for sgr in security_group_rules_param:
            rules[sgr["SecurityGroupRuleId"]] = sgr["SecurityGroupRule"]

        group_id = self._get_param("GroupId")
        self.ec2_backend.modify_security_group_rules(group_id, rules)

        return ActionResult({"Return": True})

    def update_security_group_rule_descriptions_ingress(self) -> ActionResult:
        for args in self._process_rules_from_querystring():
            # we don't need this parameter to revoke
            del args["sgrule_tags"]
            self.ec2_backend.update_security_group_rule_descriptions_ingress(**args)
        return EmptyResult()

    def update_security_group_rule_descriptions_egress(self) -> ActionResult:
        for args in self._process_rules_from_querystring():
            # we don't need this parameter to revoke
            del args["sgrule_tags"]
            self.ec2_backend.update_security_group_rule_descriptions_egress(**args)
        return EmptyResult()



"""Exceptions raised by the servicecatalogappregistry service."""

from moto.core.exceptions import JsonRESTError


class ResourceNotFoundException(JsonRESTError):
    def __init__(self, msg: str = "The specified resource does not exist.") -> None:
        super().__init__("ResourceNotFoundException", msg)


class ValidationException(JsonRESTError):
    def __init__(self) -> None:
        super().__init__(
            "ValidationException", "The request has invalid or missing parameters."
        )



from moto.utilities.utils import get_partition


def create_vector_bucket_arn(account_id: str, region: str, name: str) -> str:
    return f"arn:{get_partition(region)}:s3vectors:{region}:{account_id}:bucket/{name}"



"""Handles incoming securityhub requests, invokes methods, returns responses."""

import json
from typing import Any

from moto.core.responses import BaseResponse

from .models import SecurityHubBackend, securityhub_backends


class SecurityHubResponse(BaseResponse):
    def __init__(self) -> None:
        super().__init__(service_name="securityhub")

    @property
    def securityhub_backend(self) -> SecurityHubBackend:
        return securityhub_backends[self.current_account][self.region]

    def enable_security_hub(self) -> str:
        params = json.loads(self.body) if self.body else {}
        enable_default_standards = params.get("EnableDefaultStandards", True)
        tags = params.get("Tags", {})

        self.securityhub_backend.enable_security_hub(
            enable_default_standards=enable_default_standards,
            tags=tags,
        )
        return json.dumps({})

    def disable_security_hub(self) -> str:
        self.securityhub_backend.disable_security_hub()
        return json.dumps({})

    def describe_hub(self) -> str:
        params = json.loads(self.body) if self.body else {}
        hub_arn = params.get("HubArn")

        hub_info = self.securityhub_backend.describe_hub(hub_arn=hub_arn)
        return json.dumps(hub_info)

    def get_findings(self) -> str:
        filters = self._get_param("Filters")
        sort_criteria = self._get_param("SortCriteria")
        max_results = self._get_param("MaxResults")
        next_token = self._get_param("NextToken")

        findings, next_token = self.securityhub_backend.get_findings(
            filters=filters,
            sort_criteria=sort_criteria,
            max_results=max_results,
            next_token=next_token,
        )

        response = {"Findings": findings, "NextToken": next_token}
        return json.dumps(response)

    def batch_import_findings(self) -> str:
        raw_body = self.body
        if isinstance(raw_body, bytes):
            raw_body = raw_body.decode("utf-8")
        body = json.loads(raw_body)

        findings = body.get("Findings", [])

        failed_count, success_count, failed_findings = (
            self.securityhub_backend.batch_import_findings(
                findings=findings,
            )
        )

        return json.dumps(
            {
                "FailedCount": failed_count,
                "FailedFindings": [
                    {
                        "ErrorCode": finding.get("ErrorCode"),
                        "ErrorMessage": finding.get("ErrorMessage"),
                        "Id": finding.get("Id"),
                    }
                    for finding in failed_findings
                ],
                "SuccessCount": success_count,
            }
        )

    def enable_organization_admin_account(self) -> str:
        params = json.loads(self.body)
        admin_account_id = params.get("AdminAccountId")
        self.securityhub_backend.enable_organization_admin_account(
            admin_account_id=admin_account_id,
        )
        return json.dumps({})

    def update_organization_configuration(self) -> str:
        params = json.loads(self.body)
        auto_enable = params.get("AutoEnable")
        auto_enable_standards = params.get("AutoEnableStandards")
        organization_configuration = params.get("OrganizationConfiguration")
        self.securityhub_backend.update_organization_configuration(
            auto_enable=auto_enable,
            auto_enable_standards=auto_enable_standards,
            organization_configuration=organization_configuration,
        )
        return json.dumps({})

    def get_administrator_account(self) -> str:
        administrator = self.securityhub_backend.get_administrator_account()

        return json.dumps(administrator)

    def describe_organization_configuration(self) -> str:
        response = self.securityhub_backend.describe_organization_configuration()
        return json.dumps(dict(response))

    def create_members(self) -> str:
        params = json.loads(self.body) if self.body else {}
        account_details = params.get("AccountDetails", [])
        unprocessed_accounts = self.securityhub_backend.create_members(
            account_details=account_details,
        )
        return json.dumps({"UnprocessedAccounts": unprocessed_accounts})

    def get_members(self) -> str:
        params = json.loads(self.body) if self.body else {}
        account_ids = params.get("AccountIds", [])
        members, unprocessed_accounts = self.securityhub_backend.get_members(
            account_ids=account_ids,
        )
        return json.dumps(
            {"Members": members, "UnprocessedAccounts": unprocessed_accounts}
        )

    def list_members(self) -> str:
        only_associated = self._get_param("OnlyAssociated")
        max_results = self._get_param("MaxResults")
        if max_results is not None:
            max_results = int(max_results)
        next_token = self._get_param("NextToken")

        members, next_token = self.securityhub_backend.list_members(
            only_associated=only_associated,
            max_results=max_results,
            next_token=next_token,
        )
        response: dict[str, Any] = {"Members": members}
        if next_token:
            response["NextToken"] = next_token
        return json.dumps(response)

    def get_master_account(self) -> str:
        master = self.securityhub_backend.get_master_account()
        return json.dumps(master)



from moto.utilities.utils import get_partition


def make_arn_for_load_balancer(
    account_id: str, lb_id: str, name: str, region_name: str
) -> str:
    return f"arn:{get_partition(region_name)}:elasticloadbalancing:{region_name}:{account_id}:loadbalancer/app/{name}/{lb_id}"


def make_arn_for_target_group(
    account_id: str, tg_id: str, name: str, region_name: str
) -> str:
    return f"arn:{get_partition(region_name)}:elasticloadbalancing:{region_name}:{account_id}:targetgroup/{name}/{tg_id}"



"""Exceptions raised by the dsql service."""

from moto.core.exceptions import ServiceException


class ResourceNotFoundException(ServiceException):
    code = "ResourceNotFoundException"

    def __init__(self, arn: str, resource_id: str, resource_type: str):
        self.resource_id = resource_id
        self.resource_type = resource_type
        message = (
            f"The resource with ARN {arn} doesn't exist. Verify the ARN and try again."
        )
        super().__init__(message)



import functools
from collections.abc import Callable, Generator
from contextlib import contextmanager
from typing import (
    TYPE_CHECKING,
    ClassVar,
    Optional,
    TypeVar,
)
from urllib.parse import urlparse

import requests

from moto import settings
from moto.utilities.utils import get_partition

if TYPE_CHECKING:
    from typing_extensions import ParamSpec

    P = ParamSpec("P")

T = TypeVar("T")


class ActionAuthenticatorMixin:
    request_count: ClassVar[int] = 0

    PUBLIC_OPERATIONS = [
        "AWSCognitoIdentityService.GetId",
        "AWSCognitoIdentityService.GetOpenIdToken",
        "AWSCognitoIdentityProviderService.ConfirmSignUp",
        "AWSCognitoIdentityProviderService.GetUser",
        "AWSCognitoIdentityProviderService.ForgotPassword",
        "AWSCognitoIdentityProviderService.InitiateAuth",
        "AWSCognitoIdentityProviderService.SignUp",
    ]

    def _authenticate_and_authorize_action(
        self, iam_request_cls: type, resource: str = "*"
    ) -> None:
        if (
            ActionAuthenticatorMixin.request_count
            >= settings.INITIAL_NO_AUTH_ACTION_COUNT
        ):
            if (
                self.headers.get("X-Amz-Target")  # type: ignore[attr-defined]
                in ActionAuthenticatorMixin.PUBLIC_OPERATIONS
            ):
                return
            parsed_url = urlparse(self.uri)  # type: ignore[attr-defined]
            path = parsed_url.path
            if parsed_url.query:
                path += "?" + parsed_url.query
            iam_request = iam_request_cls(
                account_id=self.current_account,  # type: ignore[attr-defined]
                method=self.method,  # type: ignore[attr-defined]
                path=path,
                data=self.data,  # type: ignore[attr-defined]
                body=self.raw_body,  # type: ignore[attr-defined]
                headers=self.headers,  # type: ignore[attr-defined]
                action=self._get_action(),  # type: ignore[attr-defined]
            )
            iam_request.check_signature()
            iam_request.check_action_permitted(resource)
        else:
            ActionAuthenticatorMixin.request_count += 1

    def _authenticate_and_authorize_normal_action(self, resource: str = "*") -> None:
        from moto.iam.access_control import IAMRequest

        self._authenticate_and_authorize_action(IAMRequest, resource)

    def _authenticate_and_authorize_s3_action(
        self, bucket_name: Optional[str] = None, key_name: Optional[str] = None
    ) -> None:
        arn = f"{bucket_name or '*'}/{key_name}" if key_name else (bucket_name or "*")
        resource = f"arn:{get_partition(self.region)}:s3:::{arn}"  # type: ignore[attr-defined]

        from moto.iam.access_control import S3IAMRequest

        self._authenticate_and_authorize_action(S3IAMRequest, resource)

    @staticmethod
    def set_initial_no_auth_action_count(
        initial_no_auth_action_count: int,
    ) -> "Callable[[Callable[P, T]], Callable[P, T]]":
        _test_server_mode_endpoint = settings.test_server_mode_endpoint()

        def decorator(function: "Callable[P, T]") -> "Callable[P, T]":
            def wrapper(*args: "P.args", **kwargs: "P.kwargs") -> T:
                if settings.TEST_SERVER_MODE:
                    response = requests.post(
                        f"{_test_server_mode_endpoint}/moto-api/reset-auth",
                        data=str(initial_no_auth_action_count).encode("utf-8"),
                    )
                    original_initial_no_auth_action_count = response.json()[
                        "PREVIOUS_INITIAL_NO_AUTH_ACTION_COUNT"
                    ]
                else:
                    original_initial_no_auth_action_count = (
                        settings.INITIAL_NO_AUTH_ACTION_COUNT
                    )
                    original_request_count = ActionAuthenticatorMixin.request_count
                    settings.INITIAL_NO_AUTH_ACTION_COUNT = initial_no_auth_action_count
                    ActionAuthenticatorMixin.request_count = 0
                try:
                    result = function(*args, **kwargs)
                finally:
                    if settings.TEST_SERVER_MODE:
                        requests.post(
                            f"{_test_server_mode_endpoint}/moto-api/reset-auth",
                            data=str(original_initial_no_auth_action_count).encode(
                                "utf-8"
                            ),
                        )
                    else:
                        ActionAuthenticatorMixin.request_count = original_request_count
                        settings.INITIAL_NO_AUTH_ACTION_COUNT = (
                            original_initial_no_auth_action_count
                        )
                return result

            functools.update_wrapper(wrapper, function)
            wrapper.__wrapped__ = function  # type: ignore[attr-defined]
            return wrapper

        return decorator


@contextmanager
def enable_iam_authentication() -> Generator[None, None, None]:
    """Make it so that all fixtures and tests run with a simulation of IAM permissions."""

    old_initial_no_auth_action_count = settings.INITIAL_NO_AUTH_ACTION_COUNT
    old_request_count = ActionAuthenticatorMixin.request_count
    settings.INITIAL_NO_AUTH_ACTION_COUNT = 0
    ActionAuthenticatorMixin.request_count = 0
    try:
        yield
    finally:
        settings.INITIAL_NO_AUTH_ACTION_COUNT = old_initial_no_auth_action_count
        ActionAuthenticatorMixin.request_count = old_request_count


@contextmanager
def disable_iam_authentication() -> Generator[None, None, None]:
    """Inverse of enable()."""

    old_initial_no_auth_action_count = settings.INITIAL_NO_AUTH_ACTION_COUNT
    old_request_count = ActionAuthenticatorMixin.request_count
    settings.INITIAL_NO_AUTH_ACTION_COUNT = float("inf")
    ActionAuthenticatorMixin.request_count = 0
    try:
        yield
    finally:
        settings.INITIAL_NO_AUTH_ACTION_COUNT = old_initial_no_auth_action_count
        ActionAuthenticatorMixin.request_count = old_request_count



from __future__ import annotations

import random
from datetime import datetime
from typing import Any, Optional

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.moto_api._internal import mock_random
from moto.utilities.tagging_service import TaggingService

FAKE_VPC_ID = "vpc-0123456789abcdef0"


class Application(BaseModel):
    def __init__(
        self,
        account_id: str,
        region_name: str,
        application_name: str,
        application_description: Optional[str],
        runtime_environment: str,
        service_execution_role: str,
        application_configuration: Optional[dict[str, Any]],
        cloud_watch_logging_options: Optional[list[dict[str, str]]],
        application_mode: Optional[str],
    ):
        self.account_id = account_id
        self.region_name = region_name
        self.application_name = application_name
        self.application_description = application_description
        self.runtime_environment = runtime_environment
        self.service_execution_role = service_execution_role
        self.application_mode = application_mode

        self.app_config_description = (
            self._generate_app_config_description(application_configuration)
            if application_configuration
            else None
        )
        self.cloud_watch_logging_description = self._generate_logging_options(
            cloud_watch_logging_options
        )

        self.application_arn = self._generate_arn()
        self.application_status = "STARTING"
        self.application_version_id = 1
        self.creation_date_time = datetime.now().isoformat()
        self.last_updated_date_time = datetime.now().isoformat()
        self.conditional_token = str(mock_random.uuid4()).replace("-", "")

    def _generate_arn(self) -> str:
        return f"arn:aws:kinesisanalytics:{self.region_name}:{self.account_id}:application/{self.application_name}"

    def _generate_logging_options(
        self, cloud_watch_logging_options: Optional[list[dict[str, str]]]
    ) -> list[dict[str, str]] | None:
        cloud_watch_logging_option_descriptions = []
        option_id = f"{str(random.randint(1, 100))}.1"

        # Leaving out RoleARN since it is provided only sometimes for backwards
        # compatibility. Current API versions do not have the resource-level
        # role.
        if cloud_watch_logging_options:
            for i in cloud_watch_logging_options:
                cloud_watch_logging_option_descriptions.append(
                    {
                        "CloudWatchLoggingOptionId": option_id,
                        "LogStreamARN": i["LogStreamARN"],
                    }
                )
            return cloud_watch_logging_option_descriptions
        else:
            return None

    # The app_config description does not include
    # - "SqlApplicationConfigurationDescription" (discontinued)
    # - "RunConfigurationDescription" (which requires start_application)
    def _generate_app_config_description(
        self, app_config: dict[str, Any]
    ) -> dict[str, Any]:
        # Keys that do not have extra values in the description besides renamed keys
        UPDATABLE_APP_CONFIG_TOP_LEVEL_KEYS = {
            "EnvironmentProperties": "EnvironmentPropertyDescriptions",
            "ApplicationSnapshotConfiguration": "ApplicationSnapshotConfigurationDescription",
            "ApplicationSystemRollbackConfiguration": "ApplicationSystemRollbackConfigurationDescription",
            "ZeppelinApplicationConfiguration": "ZeppelinApplicationConfigurationDescription",
        }

        APP_CONFIG_SUBFIELD_KEYS = {
            "PropertyGroups": "PropertyGroupDescriptions",
            "MonitoringConfiguration": "MonitoringConfigurationDescription",
            "CatalogConfiguration": "CatalogConfigurationDescription",
            "DeployAsApplicationConfiguration": "DeployAsApplicationConfigurationDescription",
            "S3ContentLocation": "S3ContentLocationDescription",
            "CustomArtifactsConfiguration": "CustomArtifactsConfigurationDescription",
            "GlueDataCatalogConfiguration": "GlueDataCatalogConfigurationDescription",
            "MavenReference": "MavenReferenceDescription",
        }

        app_config_description = {}
        if app_config:
            if "FlinkApplicationConfiguration" in app_config:
                app_config_description["FlinkApplicationConfigurationDescription"] = (
                    self.__generate_flink_app_description(app_config)
                )

            for old_key, new_key in UPDATABLE_APP_CONFIG_TOP_LEVEL_KEYS.items():
                if old_key in app_config:
                    app_config_description[new_key] = self.__update_keys(
                        app_config[old_key], APP_CONFIG_SUBFIELD_KEYS
                    )

            app_code_config = app_config.get("ApplicationCodeConfiguration")
            if app_code_config:
                new_key = "ApplicationCodeConfigurationDescription"

                # S3ContentLocation has a different value, so keeping it
                # separate from APP_CONFIG_SUBFIELD_KEYS
                app_code_config_keys = {
                    "S3ContentLocation": "S3ApplicationCodeLocationDescription",
                    "CodeContent": "CodeContentDescription",
                }

                app_config_description[new_key] = self.__update_keys(
                    app_code_config, app_code_config_keys
                )

                if app_code_config["CodeContentType"] == "ZIPFILE":
                    app_config_description[new_key]["CodeContentDescription"][
                        "CodeMD5"
                    ] = "fakechecksum"
                    app_config_description[new_key]["CodeContentDescription"][
                        "CodeSize"
                    ] = 123

            if "VpcConfigurations" in app_config:
                app_config_description["VpcConfigurationDescriptions"] = app_config[
                    "VpcConfigurations"
                ]
                for index, vpc_config in enumerate(
                    app_config_description["VpcConfigurationDescriptions"]
                ):
                    vpc_config["VpcConfigurationId"] = str(index + 1.1)  # type: ignore[index]
                    # FAKE_VPC_ID hardcoded, not a value from the parameters
                    vpc_config["VpcId"] = FAKE_VPC_ID  # type: ignore[index]

        return app_config_description

    def __generate_flink_app_description(
        self, app_config: dict[str, Any]
    ) -> dict[str, Any]:
        flink_config_description = {}
        flink_config = app_config.get("FlinkApplicationConfiguration")
        if flink_config and isinstance(flink_config, dict):
            checkpoint_config = flink_config.get("CheckpointConfiguration")
            if checkpoint_config and isinstance(checkpoint_config, dict):
                if checkpoint_config.get("ConfigurationType") == "DEFAULT":
                    flink_config_description["CheckpointConfigurationDescription"] = {
                        "ConfigurationType": "DEFAULT",
                        "CheckpointingEnabled": True,
                        "CheckpointInterval": 60000,
                        "MinPauseBetweenCheckpoints": 5000,
                    }
                elif checkpoint_config.get("ConfigurationType") == "CUSTOM":
                    flink_config_description["CheckpointConfigurationDescription"] = {
                        "ConfigurationType": "CUSTOM",
                        "CheckpointingEnabled": checkpoint_config.get(
                            "CheckpointingEnabled", True
                        ),
                        "CheckpointInterval": checkpoint_config.get(
                            "CheckpointInterval", 60000
                        ),
                        "MinPauseBetweenCheckpoints": checkpoint_config.get(
                            "MinPauseBetweenCheckpoints", 5000
                        ),
                    }

            monitoring_config = flink_config.get("MonitoringConfiguration")
            if monitoring_config and isinstance(monitoring_config, dict):
                if monitoring_config.get("ConfigurationType") == "DEFAULT":
                    flink_config_description["MonitoringConfigurationDescription"] = {
                        "ConfigurationType": "DEFAULT",
                        "MetricsLevel": "APPLICATION",
                        "LogLevel": "INFO",
                    }
                elif monitoring_config.get("ConfigurationType") == "CUSTOM":
                    flink_config_description["MonitoringConfigurationDescription"] = {
                        "ConfigurationType": "CUSTOM",
                        "MetricsLevel": monitoring_config.get(
                            "MetricsLevel", "APPLICATION"
                        ),
                        "LogLevel": monitoring_config.get("LogLevel", "INFO"),
                    }
            parallel_config = flink_config.get("ParallelismConfiguration")
            if monitoring_config and isinstance(parallel_config, dict):
                if parallel_config.get("ConfigurationType") == "DEFAULT":
                    flink_config_description["ParallelismConfigurationDescription"] = {
                        "ConfigurationType": "DEFAULT",
                        "Parallelism": 1,
                        "ParallelismPerKPU": 1,
                        "AutoScalingEnabled": False,
                        "CurrentParallelism": 1,
                    }
                elif parallel_config.get("ConfigurationType") == "CUSTOM":
                    flink_config_description["ParallelismConfigurationDescription"] = {
                        "ConfigurationType": "CUSTOM",
                        "Parallelism": parallel_config.get("Parallelism", 1),
                        "ParallelismPerKPU": parallel_config.get(
                            "ParallelismPerKPU", 1
                        ),
                        "AutoScalingEnabled": parallel_config.get(
                            "AutoScalingEnabled", False
                        ),
                        "CurrentParallelism": parallel_config.get("Parallelism", 1),
                    }
        return flink_config_description

    def __update_keys(self, old_dict: Any, key_map: dict[str, str]) -> Any:
        if not isinstance(old_dict, dict):
            return old_dict

        updated_dict = {}
        for old_key, value in old_dict.items():
            # Check if the current key is in key_map, else keep old_key
            new_key = key_map.get(old_key, old_key)

            if isinstance(value, dict):
                updated_dict[new_key] = self.__update_keys(value, key_map)
            elif isinstance(value, list):
                updated_dict[new_key] = [
                    self.__update_keys(list_item, key_map) for list_item in value
                ]
            else:
                updated_dict[new_key] = value
        return updated_dict


class KinesisAnalyticsV2Backend(BaseBackend):
    """Implementation of KinesisAnalyticsV2 APIs."""

    def __init__(self, region_name: str, account_id: str) -> None:
        super().__init__(region_name, account_id)
        self.applications: dict[str, Application] = {}
        self.tagger = TaggingService(
            tag_name="Tags", key_name="Key", value_name="Value"
        )

    def create_application(
        self,
        application_name: str,
        application_description: Optional[str],
        runtime_environment: str,
        service_execution_role: str,
        application_configuration: Optional[dict[str, Any]],
        cloud_watch_logging_options: Optional[list[dict[str, str]]],
        tags: Optional[list[dict[str, str]]],
        application_mode: Optional[str],
    ) -> dict[str, Any]:
        app = Application(
            account_id=self.account_id,
            region_name=self.region_name,
            application_name=application_name,
            application_description=application_description,
            runtime_environment=runtime_environment,
            service_execution_role=service_execution_role,
            application_configuration=application_configuration,
            cloud_watch_logging_options=cloud_watch_logging_options,
            application_mode=application_mode,
        )

        self.applications[application_name] = app

        if tags:
            self.tag_resource(resource_arn=app.application_arn, tags=tags)
        return {
            "ApplicationARN": app.application_arn,
            "ApplicationDescription": app.application_description,
            "RuntimeEnvironment": app.runtime_environment,
            "ServiceExecutionRole": app.service_execution_role,
            "ApplicationStatus": app.application_status,
            "ApplicationVersionId": app.application_version_id,
            "CreateTimestamp": app.creation_date_time,
            "LastUpdateTimestamp": app.last_updated_date_time,
            "ApplicationConfigurationDescription": app.app_config_description,
            "CloudWatchLoggingOptionDescriptions": app.cloud_watch_logging_description,
            "ApplicationMaintenanceConfigurationDescription": {
                "ApplicationMaintenanceWindowStartTime": "06:00",
                "ApplicationMaintenanceWindowEndTime": "14:00",
            },
            "ApplicationVersionCreateTimestamp": str(app.creation_date_time),
            "ConditionalToken": app.conditional_token,
            "ApplicationMode": app.application_mode,
        }

    def tag_resource(self, resource_arn: str, tags: list[dict[str, str]]) -> None:
        self.tagger.tag_resource(resource_arn, tags)

    def list_tags_for_resource(self, resource_arn: str) -> list[dict[str, str]]:
        return self.tagger.list_tags_for_resource(resource_arn)["Tags"]

    def describe_application(
        self,
        application_name: str,
    ) -> dict[str, Any]:
        app = self.applications[application_name]
        return {
            "ApplicationARN": app.application_arn,
            "ApplicationDescription": app.application_description,
            "RuntimeEnvironment": app.runtime_environment,
            "ServiceExecutionRole": app.service_execution_role,
            "ApplicationStatus": app.application_status,
            "ApplicationVersionId": app.application_version_id,
            "CreateTimestamp": app.creation_date_time,
            "LastUpdateTimestamp": app.last_updated_date_time,
            "ApplicationConfigurationDescription": app.app_config_description,
            "CloudWatchLoggingOptionDescriptions": app.cloud_watch_logging_description,
            "ApplicationMaintenanceConfigurationDescription": {
                "ApplicationMaintenanceWindowStartTime": "06:00",
                "ApplicationMaintenanceWindowEndTime": "14:00",
            },
            "ApplicationVersionCreateTimestamp": str(app.creation_date_time),
            "ConditionalToken": app.conditional_token,
            "ApplicationMode": app.application_mode,
        }

    def list_applications(self) -> list[dict[str, Any]]:
        application_summaries = [
            {
                "ApplicationName": app.application_name,
                "ApplicationARN": app.application_arn,
                "ApplicationStatus": app.application_status,
                "ApplicationVersionId": app.application_version_id,
                "RuntimeEnvironment": app.runtime_environment,
                "ApplicationMode": app.application_mode,
            }
            for app in self.applications.values()
        ]
        return application_summaries


kinesisanalyticsv2_backends = BackendDict(
    KinesisAnalyticsV2Backend, "kinesisanalyticsv2"
)



from moto.core.exceptions import ServiceException


class ElastiCacheException(ServiceException):
    pass


class PasswordTooShort(ElastiCacheException):
    code = "InvalidParameterValue"
    message = "Passwords length must be between 16-128 characters."


class PasswordRequired(ElastiCacheException):
    code = "InvalidParameterValue"
    message = "No password was provided. If you want to create/update the user without password, please use the NoPasswordRequired flag."


class InvalidParameterValueException(ElastiCacheException):
    code = "InvalidParameterValue"


class InvalidParameterCombinationException(ElastiCacheException):
    code = "InvalidParameterCombination"


class UserAlreadyExists(ElastiCacheException):
    code = "UserAlreadyExists"

    def __init__(self) -> None:
        super().__init__("User user1 already exists.")


class UserNotFound(ElastiCacheException):
    code = "UserNotFound"

    def __init__(self, user_id: str):
        super().__init__(f"User {user_id} not found.")


class CacheClusterAlreadyExists(ElastiCacheException):
    code = "CacheClusterAlreadyExists"

    def __init__(self, cache_cluster_id: str):
        super().__init__(f"Cache cluster {cache_cluster_id} already exists.")


class CacheClusterNotFound(ElastiCacheException):
    code = "CacheClusterNotFound"

    def __init__(self, cache_cluster_id: str):
        super().__init__(f"Cache cluster {cache_cluster_id} not found.")


class CacheSubnetGroupAlreadyExists(ElastiCacheException):
    code = "CacheSubnetGroupAlreadyExists"

    def __init__(self, cache_subnet_group_name: str):
        super().__init__(f"CacheSubnetGroup {cache_subnet_group_name} already exists.")


class CacheSubnetGroupNotFound(ElastiCacheException):
    code = "CacheSubnetGroupNotFoundFault"

    def __init__(self, cache_subnet_group_name: str):
        super().__init__(f"CacheSubnetGroup {cache_subnet_group_name} not found.")


class InvalidARNFault(ElastiCacheException):
    code = "InvalidARN"

    def __init__(self, arn: str):
        super().__init__(f"ARN {arn} is invalid.")


class InvalidSubnet(ElastiCacheException):
    code = "InvalidSubnet"

    def __init__(self, subnet_id: str):
        super().__init__(f"Subnet {subnet_id} is invalid.")


class ReplicationGroupAlreadyExists(ElastiCacheException):
    code = "ReplicationGroupAlreadyExists"

    def __init__(self, replication_group_id: str):
        super().__init__(f"Replication group {replication_group_id} already exists.")


class ReplicationGroupNotFound(ElastiCacheException):
    code = "ReplicationGroupNotFoundFault"

    def __init__(self, replication_group_id: str):
        super().__init__(f"Replication group {replication_group_id} not found.")


class SnapshotNotFound(ElastiCacheException):
    code = "SnapshotNotFoundFault"

    def __init__(self, snapshot_name: str):
        super().__init__(f"Snapshot {snapshot_name} not found.")


class SnapshotAlreadyExists(ElastiCacheException):
    code = "SnapshotAlreadyExistsFault"

    def __init__(self, snapshot_name: str):
        super().__init__(f"Snapshot {snapshot_name} already exists.")



"""Handles incoming account requests, invokes methods, returns responses."""

from moto.core.responses import ActionResult, BaseResponse, EmptyResult

from .exceptions import UnknownContactType
from .models import ALLOWED_CONTACT_TYPES, account_backends


class AccountResponse(BaseResponse):
    def __init__(self) -> None:
        super().__init__(service_name="account")

    def put_alternate_contact(self) -> ActionResult:
        account_id = self._get_account_id()
        alternate_contact_type = self._get_contact_type(account_id)
        email_address = self._get_param("EmailAddress")
        name = self._get_param("Name")
        phone_number = self._get_param("PhoneNumber")
        title = self._get_param("Title")

        backend = account_backends[account_id or self.current_account][self.partition]

        backend.put_alternate_contact(
            alternate_contact_type=alternate_contact_type,
            email_address=email_address,
            name=name,
            phone_number=phone_number,
            title=title,
        )
        return EmptyResult()

    def get_alternate_contact(self) -> ActionResult:
        account_id = self._get_account_id()
        alternate_contact_type = self._get_contact_type(account_id)

        backend = account_backends[account_id][self.partition]

        contact = backend.get_alternate_contact(
            alternate_contact_type=alternate_contact_type
        )
        return ActionResult(result={"AlternateContact": contact})

    def delete_alternate_contact(self) -> EmptyResult:
        account_id = self._get_account_id()
        alternate_contact_type = self._get_contact_type(account_id)

        backend = account_backends[account_id][self.partition]

        backend.delete_alternate_contact(alternate_contact_type=alternate_contact_type)
        return EmptyResult()

    def _get_account_id(self) -> str:
        return self._get_param("AccountId") or self.current_account

    def _get_contact_type(self, account_id: str) -> str:
        alternate_contact_type = self._get_param("AlternateContactType")
        if alternate_contact_type not in ALLOWED_CONTACT_TYPES:
            from moto.sts.models import STSBackend, sts_backends

            access_key_id = self.get_access_key()
            sts_backend: STSBackend = sts_backends[account_id][self.partition]
            _, user_arn, _ = sts_backend.get_caller_identity(access_key_id, self.region)
            raise UnknownContactType(user_arn, action="GetAlternateContact")
        return alternate_contact_type



from moto.core.responses import ActionResult
from moto.ec2.exceptions import InvalidParameterCombination
from moto.ec2.utils import parse_user_data

from ._base_response import EC2BaseResponse


class LaunchTemplates(EC2BaseResponse):
    def create_launch_template(self) -> ActionResult:
        name = self._get_param("LaunchTemplateName")
        version_description = self._get_param("VersionDescription")
        tag_spec = self._parse_tag_specification()

        parsed_template_data = self._get_param("LaunchTemplateData", {})
        parsed_template_data["UserData"] = parse_user_data(
            self._get_param("LaunchTemplateData.UserData")
        )
        self.error_on_dryrun()

        if tag_spec:
            if "TagSpecifications" not in parsed_template_data:
                parsed_template_data["TagSpecifications"] = []
            converted_tag_spec = []
            for resource_type, tags in tag_spec.items():
                converted_tag_spec.append(
                    {
                        "ResourceType": resource_type,
                        "Tags": [
                            {"Key": key, "Value": value} for key, value in tags.items()
                        ],
                    }
                )

            parsed_template_data["TagSpecifications"].extend(converted_tag_spec)

        template = self.ec2_backend.create_launch_template(
            name, version_description, parsed_template_data, tag_spec
        )
        version = template.default_version()

        result = {
            "LaunchTemplate": {
                "CreateTime": version.create_time,
                "CreatedBy": f"arn:{self.partition}:iam::{self.current_account}:root",
                "DefaultVersionNumber": template.default_version_number,
                "LatestVersionNumber": version.number,
                "LaunchTemplateId": template.id,
                "LaunchTemplateName": template.name,
                "Tags": template.tags,
            },
        }

        return ActionResult(result)

    def create_launch_template_version(self) -> ActionResult:
        name = self._get_param("LaunchTemplateName")
        tmpl_id = self._get_param("LaunchTemplateId")
        if name:
            template = self.ec2_backend.get_launch_template_by_name(name)
        if tmpl_id:
            template = self.ec2_backend.get_launch_template(tmpl_id)

        version_description = self._get_param("VersionDescription")

        template_data = self._get_param("LaunchTemplateData", {})

        self.error_on_dryrun()

        version = template.create_version(template_data, version_description)

        result = {
            "LaunchTemplateVersion": {
                "CreateTime": version.create_time,
                "CreatedBy": f"arn:{self.partition}:iam::{self.current_account}:root",
                "DefaultVersion": template.is_default(version),
                "LaunchTemplateData": version.data,
                "LaunchTemplateId": template.id,
                "LaunchTemplateName": template.name,
                "VersionDescription": version.description,
                "VersionNumber": version.number,
            },
        }
        return ActionResult(result)

    def delete_launch_template(self) -> ActionResult:
        name = self._get_param("LaunchTemplateName")
        tid = self._get_param("LaunchTemplateId")

        self.error_on_dryrun()

        template = self.ec2_backend.delete_launch_template(name, tid)

        result = {
            "LaunchTemplate": {
                "DefaultVersionNumber": template.default_version_number,
                "LaunchTemplateId": template.id,
                "LaunchTemplateName": template.name,
            },
        }

        return ActionResult(result)

    def describe_launch_template_versions(self) -> ActionResult:
        name = self._get_param("LaunchTemplateName")
        template_id = self._get_param("LaunchTemplateId")

        max_results = self._get_int_param("MaxResults", 15)
        versions = self._get_param("Versions", [])
        min_version = self._get_int_param("MinVersion")
        max_version = self._get_int_param("MaxVersion")

        self.error_on_dryrun()

        ret_versions = self.ec2_backend.describe_launch_template_versions(
            template_name=name,
            template_id=template_id,
            versions=versions,
            min_version=min_version,
            max_version=max_version,
            max_results=max_results,
        )

        result = {
            "LaunchTemplateVersions": [
                {
                    "CreateTime": version.create_time,
                    "CreatedBy": f"arn:{self.partition}:iam::{self.current_account}:root",
                    "DefaultVersion": template.is_default(version),
                    "LaunchTemplateData": version.data,
                    "LaunchTemplateId": template.id,
                    "LaunchTemplateName": template.name,
                    "VersionDescription": version.description,
                    "VersionNumber": version.number,
                }
                for template, version in ret_versions
            ]
        }

        return ActionResult(result)

    def describe_launch_templates(self) -> ActionResult:
        max_results = self._get_int_param("MaxResults", 15)
        template_names = self._get_param("LaunchTemplateNames", [])
        template_ids = self._get_param("LaunchTemplateIds", [])
        filters = self._filters_from_querystring()

        self.error_on_dryrun()

        templates = self.ec2_backend.describe_launch_templates(
            template_names=template_names,
            template_ids=template_ids,
            filters=filters,
        )

        templates = templates[:max_results]

        result = {
            "LaunchTemplates": [
                {
                    "CreateTime": template.create_time,
                    "CreatedBy": f"arn:{self.partition}:iam::{self.current_account}:root",
                    "DefaultVersionNumber": template.default_version_number,
                    "LatestVersionNumber": template.latest_version_number,
                    "LaunchTemplateId": template.id,
                    "LaunchTemplateName": template.name,
                    "Tags": template.tags,
                }
                for template in templates
            ]
        }

        return ActionResult(result)

    def get_launch_template_data(self) -> ActionResult:
        instance_id = self._get_param("InstanceId")
        instance = self.ec2_backend.get_launch_template_data(instance_id)
        # Result is based on Moto's original XML jinja template, including hardcoded values.
        result = {
            "LaunchTemplateData": {
                "BlockDeviceMappings": [
                    {
                        "DeviceName": "string",
                        "Ebs": {
                            "Encrypted": device.encrypted,
                            "DeleteOnTermination": device.delete_on_termination,
                            "SnapshotId": device.snapshot_id,
                            "VolumeSize": device.size,
                            "VolumeType": device.volume_type,
                        },
                    }
                    for device_name, device in instance.block_device_mapping.items()
                ],
                "CapacityReservationSpecification": {
                    "CapacityReservationPreference": "open",
                },
                "CreditSpecification": {"CpuCredits": "standard"},
                "DisableApiStop": instance.disable_api_stop,
                "DisableApiTermination": instance.disable_api_termination,
                "EbsOptimized": instance.ebs_optimized,
                "EnclaveOptions": {"Enabled": False},
                "HibernationOptions": {"Configured": False},
                "ImageId": instance.image_id,
                "InstanceInitiatedShutdownBehavior": instance.instance_initiated_shutdown_behavior,
                "InstanceType": instance.instance_type,
                "KeyName": instance.key_name,
                "MaintenanceOptions": {"AutoRecovery": "default"},
                "MetadataOptions": {
                    "HttpTokens": "optional",
                    "HttpPutResponseHopLimit": 1,
                    "HttpEndpoint": "enabled",
                    "HttpProtocolIpv6": "disabled",
                    "InstanceMetadataTags": "disabled",
                },
                "Monitoring": {"Enabled": instance.monitored},
                "NetworkInterfaces": [
                    {
                        "AssociatePublicIpAddress": True,
                        "DeleteOnTermination": nic.delete_on_termination,
                        "Description": "",
                        "DeviceIndex": nic.device_index,
                        "Groups": [
                            {"GroupId": group.group_id}
                            for group in nic.group_set
                            if nic.group_set
                        ],
                        "InterfaceType": nic.interface_type,
                        "PrivateIpAddresses": [
                            {
                                "Primary": addr["Primary"],
                                "PrivateIpAddress": addr["PrivateIpAddress"],
                            }
                            for addr in nic.private_ip_addresses
                        ],
                        "SubnetId": nic.subnet.id,
                        "NetworkCardIndex": nic_index,
                    }
                    for nic_index, nic in instance.nics.items()
                ],
                "Placement": {
                    "AvailabilityZone": instance.availability_zone,
                    "GroupName": "",
                    "Tenancy": "default",
                },
                "PrivateDnsNameOptions": {
                    "HostnameType": "ip-name",
                    "EnableResourceNameDnsARecord": True,
                    "EnableResourceNameDnsAAAARecord": False,
                },
                "TagSpecifications": [
                    {
                        "ResourceType": "instance",
                        "Tags": [
                            {
                                "Key": tag["key"],
                                "Value": tag["value"],
                            }
                            for tag in instance.get_tags()
                        ],
                    }
                ]
                if instance.get_tags()
                else [],
            }
        }
        return ActionResult(result)

    def modify_launch_template(self) -> ActionResult:
        template_name = self._get_param("LaunchTemplateName")
        template_id = self._get_param("LaunchTemplateId")
        default_version = self._get_param("DefaultVersion")

        if template_name and template_id:
            raise InvalidParameterCombination(
                "Either provide launch template ID or launch template name to modify the template."
            )

        self.error_on_dryrun()

        template = self.ec2_backend.modify_launch_template(
            template_name=template_name,
            template_id=template_id,
            default_version=default_version,
        )

        result = {
            "LaunchTemplate": {
                "CreateTime": template.create_time,
                "CreatedBy": f"arn:{self.partition}:iam::{self.current_account}:root",
                "DefaultVersionNumber": template.default_version_number,
                "LatestVersionNumber": template.latest_version_number,
                "LaunchTemplateId": template.id,
                "LaunchTemplateName": template.name,
                "Tags": template.tags,
            },
        }
        return ActionResult(result)



"""sesv2 base URL and path."""

from .responses import SESV2Response

url_bases = [
    r"https?://email\.(.+)\.amazonaws\.com",
]


response = SESV2Response()


url_paths = {
    "{0}/v2/email/outbound-emails$": response.dispatch,
    "{0}/v2/email/contact-lists/(?P<name>[^/]+)$": response.dispatch,
    "{0}/v2/email/contact-lists/(?P<name>[^/]+)/contacts$": response.dispatch,
    "{0}/v2/email/contact-lists/(?P<name>[^/]+)/contacts/(?P<email>[^/]+)$": response.dispatch,
    "{0}/v2/email/contact-lists$": response.dispatch,
    "{0}/v2/.*$": response.dispatch,
    "{0}/v2/email/configuration-sets$": SESV2Response.dispatch,
    "{0}/v2/email/configuration-sets/(?P<ConfigurationSetName>[^/]+)$": SESV2Response.dispatch,
    "{0}/v2/email/dedicated-ip-pools$": SESV2Response.dispatch,
    "{0}/v2/email/dedicated-ip-pools/(?P<PoolName>[^/]+)$": SESV2Response.dispatch,
    "{0}/v2/email/identities$": SESV2Response.dispatch,
    "{0}/v2/email/identities/(?P<EmailIdentity>[^/]+)$": SESV2Response.dispatch,
    "{0}/v2/email/identities/(?P<EmailIdentity>[^/]+)/policies/(?P<PolicyName>[^/]+)$": SESV2Response.dispatch,
    "{0}/v2/email/identities/(?P<EmailIdentity>[^/]+)/policies$": SESV2Response.dispatch,
    "{0}/v2/email/tags$": SESV2Response.dispatch,
}



"""Exceptions raised by the codedeploy service."""

from moto.core.exceptions import JsonRESTError


class CodeDeployException(JsonRESTError):
    pass


class ApplicationDoesNotExistException(CodeDeployException):
    code = 400

    def __init__(self, message: str):
        super().__init__("ApplicationDoesNotExistException", message)


class DeploymentDoesNotExistException(CodeDeployException):
    code = 400

    def __init__(self, message: str):
        super().__init__("DeploymentDoesNotExistException", message)


class ApplicationAlreadyExistsException(CodeDeployException):
    code = 400

    def __init__(self, message: str):
        super().__init__("ApplicationAlreadyExistsException", message)


class ApplicationNameRequiredException(CodeDeployException):
    code = 400

    def __init__(self, message: str):
        super().__init__("ApplicationNameRequiredException", message)


class DeploymentGroupAlreadyExistsException(CodeDeployException):
    code = 400

    def __init__(self, message: str):
        super().__init__("DeploymentGroupAlreadyExistsException", message)


class DeploymentGroupNameRequiredException(CodeDeployException):
    code = 400

    def __init__(self, message: str):
        super().__init__("DeploymentGroupNameRequiredException", message)


class DeploymentGroupDoesNotExistException(CodeDeployException):
    code = 400

    def __init__(self, message: str):
        super().__init__("DeploymentGroupDoesNotExistException", message)



import abc
import copy
from typing import Any, Final, Optional

from moto.stepfunctions.parser.api import HistoryEventType, TaskFailedEventDetails
from moto.stepfunctions.parser.asl.component.common.error_name.failure_event import (
    FailureEvent,
    FailureEventException,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name import (
    StatesErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name_type import (
    StatesErrorNameType,
)
from moto.stepfunctions.parser.asl.component.common.query_language import (
    QueryLanguageMode,
)
from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.component.intrinsic.jsonata import (
    get_intrinsic_functions_declarations,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.eval.event.event_detail import EventDetails
from moto.stepfunctions.parser.asl.jsonata.jsonata import (
    JSONataExpression,
    VariableDeclarations,
    VariableReference,
    compose_jsonata_expression,
    eval_jsonata_expression,
    extract_jsonata_variable_references,
)
from moto.stepfunctions.parser.asl.jsonata.validations import (
    validate_jsonata_expression_output,
)
from moto.stepfunctions.parser.asl.utils.encoding import to_json_str
from moto.stepfunctions.parser.asl.utils.json_path import (
    NoSuchJsonPathError,
    extract_json,
)

JSONPATH_ROOT_PATH: Final[str] = "$"


class StringExpression(EvalComponent, abc.ABC):
    literal_value: Final[str]

    def __init__(self, literal_value: str):
        self.literal_value = literal_value

    def _field_name(self) -> Optional[str]:
        return None


class StringExpressionSimple(StringExpression, abc.ABC): ...


class StringSampler(StringExpressionSimple, abc.ABC): ...


class StringLiteral(StringExpression):
    def _eval_body(self, env: Environment) -> None:
        env.stack.append(self.literal_value)


class StringJsonPath(StringSampler):
    json_path: Final[str]

    def __init__(self, json_path: str):
        super().__init__(literal_value=json_path)
        self.json_path = json_path

    def _eval_body(self, env: Environment) -> None:
        input_value: Any = env.stack[-1]
        if self.json_path == JSONPATH_ROOT_PATH:
            output_value = input_value
        else:
            output_value = extract_json(self.json_path, input_value)
        # TODO: introduce copy on write approach
        env.stack.append(copy.deepcopy(output_value))


class StringContextPath(StringJsonPath):
    context_object_path: Final[str]

    def __init__(self, context_object_path: str):
        json_path = context_object_path[1:]
        super().__init__(json_path=json_path)
        self.context_object_path = context_object_path

    def _eval_body(self, env: Environment) -> None:
        input_value = env.states.context_object.context_object_data
        if self.json_path == JSONPATH_ROOT_PATH:
            output_value = input_value
        else:
            try:
                output_value = extract_json(self.json_path, input_value)
            except NoSuchJsonPathError:
                input_value_json_str = to_json_str(input_value)
                cause = (
                    f"The JSONPath '${self.json_path}' specified for the field '{env.next_field_name}' "
                    f"could not be found in the input '{input_value_json_str}'"
                )
                raise FailureEventException(
                    failure_event=FailureEvent(
                        env=env,
                        error_name=StatesErrorName(
                            typ=StatesErrorNameType.StatesRuntime
                        ),
                        event_type=HistoryEventType.TaskFailed,
                        event_details=EventDetails(
                            taskFailedEventDetails=TaskFailedEventDetails(
                                error=StatesErrorNameType.StatesRuntime.to_name(),
                                cause=cause,
                            )
                        ),
                    )
                )
        # TODO: introduce copy on write approach
        env.stack.append(copy.deepcopy(output_value))


class StringVariableSample(StringSampler):
    query_language_mode: Final[QueryLanguageMode]
    expression: Final[str]

    def __init__(self, query_language_mode: QueryLanguageMode, expression: str):
        super().__init__(literal_value=expression)
        self.query_language_mode = query_language_mode
        self.expression = expression

    def _eval_body(self, env: Environment) -> None:
        # Get the variables sampled in the jsonata expression.
        expression_variable_references: set[VariableReference] = (
            extract_jsonata_variable_references(self.expression)
        )
        variable_declarations_list = []
        if self.query_language_mode == QueryLanguageMode.JSONata:
            # Sample $states values into expression.
            states_variable_declarations: VariableDeclarations = (
                env.states.to_variable_declarations(
                    variable_references=expression_variable_references
                )
            )
            variable_declarations_list.append(states_variable_declarations)

        # Sample Variable store values in to expression.
        # TODO: this could be optimised by sampling only those invoked.
        variable_declarations: VariableDeclarations = (
            env.variable_store.get_variable_declarations()
        )
        variable_declarations_list.append(variable_declarations)

        rich_jsonata_expression: JSONataExpression = compose_jsonata_expression(
            final_jsonata_expression=self.expression,
            variable_declarations_list=variable_declarations_list,
        )
        result = eval_jsonata_expression(rich_jsonata_expression)
        env.stack.append(result)


class StringIntrinsicFunction(StringExpressionSimple):
    intrinsic_function_derivation: Final[str]
    function: Final[EvalComponent]

    def __init__(
        self, intrinsic_function_derivation: str, function: EvalComponent
    ) -> None:
        super().__init__(literal_value=intrinsic_function_derivation)
        self.intrinsic_function_derivation = intrinsic_function_derivation
        self.function = function

    def _eval_body(self, env: Environment) -> None:
        self.function.eval(env=env)


class StringJSONata(StringExpression):
    expression: Final[str]

    def __init__(self, expression: str):
        super().__init__(literal_value=expression)
        # TODO: check for illegal functions ($, $$, $eval)
        self.expression = expression

    def _eval_body(self, env: Environment) -> None:
        # Get the variables sampled in the jsonata expression.
        expression_variable_references: set[VariableReference] = (
            extract_jsonata_variable_references(self.expression)
        )

        # Sample declarations for used intrinsic functions. Place this at the start allowing users to
        # override these identifiers with custom variable declarations.
        functions_variable_declarations: VariableDeclarations = (
            get_intrinsic_functions_declarations(
                variable_references=expression_variable_references
            )
        )

        # Sample $states values into expression.
        states_variable_declarations: VariableDeclarations = (
            env.states.to_variable_declarations(
                variable_references=expression_variable_references
            )
        )

        # Sample Variable store values in to expression.
        # TODO: this could be optimised by sampling only those invoked.
        variable_declarations: VariableDeclarations = (
            env.variable_store.get_variable_declarations()
        )

        rich_jsonata_expression: JSONataExpression = compose_jsonata_expression(
            final_jsonata_expression=self.expression,
            variable_declarations_list=[
                functions_variable_declarations,
                states_variable_declarations,
                variable_declarations,
            ],
        )
        result = eval_jsonata_expression(rich_jsonata_expression)

        validate_jsonata_expression_output(
            env, self.expression, rich_jsonata_expression, result
        )

        env.stack.append(result)



"""ServiceCatalogBackend class with methods for supported APIs."""

import uuid
from datetime import datetime
from typing import Any, Optional

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.utilities.paginator import paginate
from moto.utilities.tagging_service import TaggingService
from moto.utilities.utils import get_partition

from .exceptions import (
    InvalidParametersException,
    ResourceNotFoundException,
)

PAGINATION_MODEL = {
    "list_portfolio_access": {
        "input_token": "page_token",
        "limit_key": "page_size",
        "limit_default": 20,
        "unique_attribute": "account_id",
    }
}


class Portfolio(BaseModel):
    """Portfolio resource."""

    def __init__(
        self,
        portfolio_id: str,
        display_name: str,
        description: str,
        provider_name: str,
        tags: list[dict[str, str]],
        region_name: str,
        account_id: str,
        backend: "ServiceCatalogBackend",
    ) -> None:
        self.id = portfolio_id
        self.display_name = display_name
        self.description = description
        self.provider_name = provider_name
        self.created_time = datetime.now()
        self.tags = tags
        self.region_name = region_name
        self.account_id = account_id
        self.backend = backend
        if tags:
            self.backend._tag_resource(self.arn, tags)

    @property
    def arn(self) -> str:
        return f"arn:{get_partition(self.region_name)}:catalog:{self.region_name}:{self.account_id}:portfolio/{self.id}"

    def to_dict(self) -> dict[str, str]:
        return {
            "Id": self.id,
            "ARN": self.arn,
            "DisplayName": self.display_name,
            "Description": self.description,
            "CreatedTime": self.created_time.isoformat(),
            "ProviderName": self.provider_name,
        }


class Product(BaseModel):
    """Service Catalog Product resource."""

    def __init__(
        self,
        name: str,
        owner: str,
        description: Optional[str],
        distributor: Optional[str],
        support_description: Optional[str],
        support_email: Optional[str],
        support_url: Optional[str],
        product_type: str,
        tags: Optional[list[dict[str, str]]],
        provisioning_artifact_parameters: Optional[dict[str, Any]],
        source_connection: Optional[dict[str, Any]],
        accept_language: Optional[str],
        region_name: str,
        account_id: str,
        backend: "ServiceCatalogBackend",
    ) -> None:
        self.id = str(uuid.uuid4())
        self.name = name
        self.owner = owner
        self.description = description
        self.distributor = distributor
        self.support_description = support_description
        self.support_email = support_email
        self.support_url = support_url
        self.product_type = product_type
        self.tags = tags or []
        self.provisioning_artifact_parameters = provisioning_artifact_parameters or {}
        self.source_connection = source_connection or {}
        self.accept_language = accept_language
        self.created_time = datetime.now()
        self.region_name = region_name
        self.account_id = account_id
        self.backend = backend
        if tags:
            self.backend._tag_resource(self.arn, tags)

    @property
    def arn(self) -> str:
        return f"arn:{get_partition(self.region_name)}:catalog:{self.region_name}:{self.account_id}:product/{self.id}"

    def to_dict(self) -> dict[str, Any]:
        product_view_summary = {
            "Id": self.id,  # ?
            "ProductId": self.id,  # ?
            "Name": self.name,
            "Owner": self.owner,
            "Type": self.product_type,
            "Distributor": self.distributor,
            "HasDefaultPath": False,
            "ShortDescription": self.description,
            "SupportDescription": self.support_description,
            "SupportEmail": self.support_email,
            "SupportUrl": self.support_url,
        }

        return product_view_summary


class ServiceCatalogBackend(BaseBackend):
    """Implementation of ServiceCatalog APIs."""

    def __init__(self, region_name: str, account_id: str) -> None:
        super().__init__(region_name, account_id)
        self.portfolio_access: dict[str, list[str]] = {}
        self.portfolios: dict[str, Portfolio] = {}
        self.idempotency_tokens: dict[str, str] = {}
        self.portfolio_share_tokens: dict[str, list[str]] = {}
        self.products: dict[str, Product] = {}
        self.tagger = TaggingService()

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_portfolio_access(
        self,
        accept_language: Optional[str],
        portfolio_id: str,
        organization_parent_id: Optional[str],
        page_token: Optional[str],
        page_size: Optional[int] = None,
    ) -> list[dict[str, str]]:
        # TODO: Implement organization_parent_id and accept_language

        account_ids = self.portfolio_access.get(portfolio_id, [])
        return [{"account_id": account_id} for account_id in account_ids]

    def delete_portfolio(self, accept_language: Optional[str], id: str) -> None:
        # TODO: Implement accept_language

        if id in self.portfolio_access:
            del self.portfolio_access[id]

        if id in self.portfolios:
            del self.portfolios[id]

        return None

    def delete_portfolio_share(
        self,
        accept_language: Optional[str],
        portfolio_id: str,
        account_id: Optional[str],
        organization_node: Optional[dict[str, str]],
    ) -> Optional[str]:
        # TODO: Implement accept_language

        if (
            portfolio_id in self.portfolio_access
            and account_id in self.portfolio_access[portfolio_id]
        ):
            self.portfolio_access[portfolio_id].remove(account_id)

        portfolio_share_token = None
        if organization_node:
            org_type = organization_node.get("Type", "")
            org_value = organization_node.get("Value", "")

            # Arbitrary naming for the portfolio share token
            portfolio_share_token = f"share-{portfolio_id}-{org_type}-{org_value}"

            if portfolio_id in self.portfolio_share_tokens:
                tokens_to_remove = []

                for token in self.portfolio_share_tokens[portfolio_id]:
                    if org_type in token and org_value in token:
                        tokens_to_remove.append(token)

                for token in tokens_to_remove:
                    if token in self.portfolio_share_tokens[portfolio_id]:
                        self.portfolio_share_tokens[portfolio_id].remove(token)

        return portfolio_share_token

    def create_portfolio(
        self,
        accept_language: Optional[str],
        display_name: str,
        description: Optional[str],
        provider_name: str,
        tags: Optional[list[dict[str, str]]],
        idempotency_token: Optional[str],
    ) -> tuple[dict[str, str], list[dict[str, str]]]:
        # TODO: Implement accept_language

        if idempotency_token and idempotency_token in self.idempotency_tokens:
            portfolio_id = self.idempotency_tokens[idempotency_token]
            portfolio = self.portfolios[portfolio_id]
            return portfolio.to_dict(), portfolio.tags

        portfolio_id = str(uuid.uuid4())

        portfolio = Portfolio(
            portfolio_id=portfolio_id,
            display_name=display_name,
            description=description or "",
            provider_name=provider_name,
            tags=tags or [],
            region_name=self.region_name,
            account_id=self.account_id,
            backend=self,
        )

        self.portfolios[portfolio_id] = portfolio

        if idempotency_token:
            self.idempotency_tokens[idempotency_token] = portfolio_id

        self.portfolio_access[portfolio_id] = []

        return portfolio.to_dict(), portfolio.tags

    def create_portfolio_share(
        self,
        accept_language: Optional[str],
        portfolio_id: str,
        account_id: Optional[str],
        organization_node: Optional[dict[str, str]],
        share_tag_options: bool,
        share_principals: bool,
    ) -> Optional[str]:
        # TODO: Implement accept_language

        if portfolio_id not in self.portfolios:
            return None

        if account_id:
            if portfolio_id not in self.portfolio_access:
                self.portfolio_access[portfolio_id] = []

            if account_id not in self.portfolio_access[portfolio_id]:
                self.portfolio_access[portfolio_id].append(account_id)

            return None

        portfolio_share_token = None
        if organization_node:
            org_type = organization_node.get("Type", "")
            org_value = organization_node.get("Value", "")

            # Arbitrary org for the portfolio share token
            portfolio_share_token = f"share-{portfolio_id}-{org_type}-{org_value}"

            if share_tag_options:
                portfolio_share_token += "-tags"
            if share_principals:
                portfolio_share_token += "-principals"

            if portfolio_id not in self.portfolio_share_tokens:
                self.portfolio_share_tokens[portfolio_id] = []

            if portfolio_share_token not in self.portfolio_share_tokens[portfolio_id]:
                self.portfolio_share_tokens[portfolio_id].append(portfolio_share_token)

        return portfolio_share_token

    def list_portfolios(
        self,
        accept_language: Optional[str] = None,
        page_token: Optional[str] = None,
        page_size: Optional[int] = None,
    ) -> tuple[list[dict[str, str]], Optional[str]]:
        """TODO: Implement pagination and accept_language"""
        portfolio_details = [
            portfolio.to_dict() for portfolio in self.portfolios.values()
        ]
        return portfolio_details, None

    def describe_portfolio_shares(
        self,
        portfolio_id: str,
        type: str,
        page_token: Optional[str] = None,
        page_size: Optional[int] = None,
    ) -> tuple[Optional[str], list[dict[str, Any]]]:
        """TODO: Implement pagination"""

        portfolio_share_details = []

        if portfolio_id not in self.portfolios:
            return None, []

        if type == "ACCOUNT":
            account_ids = self.portfolio_access.get(portfolio_id, [])
            for account_id in account_ids:
                portfolio_share_details.append(
                    {
                        "PrincipalId": account_id,
                        "Type": "ACCOUNT",
                        "Accepted": True,
                        "ShareTagOptions": False,
                        "SharePrincipals": False,
                    }
                )

        elif type == "ORGANIZATION":
            tokens = self.portfolio_share_tokens.get(portfolio_id, [])

            for token in tokens:
                if "ORGANIZATION" in token and "-o-" in token:
                    org_id_start = token.find("-ORGANIZATION-") + len("-ORGANIZATION-")

                    if "-tags" in token:
                        org_id_end = token.find("-tags", org_id_start)
                    elif "-principals" in token:
                        org_id_end = token.find("-principals", org_id_start)
                    else:
                        org_id_end = len(token)

                    org_id = token[org_id_start:org_id_end]

                    portfolio_share_details.append(
                        {
                            "PrincipalId": org_id,
                            "Type": "ORGANIZATION",
                            "Accepted": True,
                            "ShareTagOptions": "tags" in token,
                            "SharePrincipals": "principals" in token,
                        }
                    )

        return None, portfolio_share_details

    def describe_portfolio(
        self, accept_language: Optional[str], id: str
    ) -> tuple[
        dict[str, Any], list[dict[str, str]], list[dict[str, Any]], list[dict[str, str]]
    ]:
        # TODO: Implement accept_language

        if id not in self.portfolios:
            return {}, [], [], []

        portfolio = self.portfolios[id]
        portfolio_detail = portfolio.to_dict()

        tags = portfolio.tags

        tag_options: list[dict[str, Any]] = []
        budgets: list[dict[str, Any]] = []

        return portfolio_detail, tags, tag_options, budgets

    def create_product(
        self,
        name: str,
        owner: str,
        description: Optional[str],
        distributor: Optional[str],
        support_description: Optional[str],
        support_email: Optional[str],
        support_url: Optional[str],
        product_type: str,
        tags: list[dict[str, str]],
        provisioning_artifact_parameters: Optional[dict[str, Any]],
        idempotency_token: Optional[str] = None,
        source_connection: Optional[dict[str, Any]] = None,
        accept_language: Optional[str] = None,
    ) -> Product:
        token = idempotency_token or str(uuid.uuid4())
        existing_id = self.idempotency_tokens.get(token)
        if existing_id:
            return self.products[existing_id]

        product = Product(
            name=name,
            owner=owner,
            description=description,
            distributor=distributor,
            support_description=support_description,
            support_email=support_email,
            support_url=support_url,
            product_type=product_type,
            tags=tags,
            provisioning_artifact_parameters=provisioning_artifact_parameters,
            source_connection=source_connection,
            accept_language=accept_language,
            region_name=self.region_name,
            account_id=self.account_id,
            backend=self,
        )

        self.products[product.id] = product
        self.idempotency_tokens[token] = product.id

        return product

    def describe_product(
        self, accept_language: Optional[str], id: str, name: str
    ) -> Product:
        if not id and not name:
            raise InvalidParametersException("Either Id or Name must be specified.")

        if id:
            product = self.products.get(id)
            if not product:
                raise ResourceNotFoundException(f"Product with Id '{id}' not found.")
        else:
            product = self.lookup_by_name(name)
            if not product:
                raise ResourceNotFoundException(
                    f"Product with Name '{name}' not found."
                )

        return product

    def lookup_by_name(self, name: str) -> Optional[Product]:
        for product in self.products.values():
            if name == product.name:
                return product

        return None

    def delete_product(self, accept_language: Optional[str], id: str) -> None:
        if id in self.products:
            del self.products[id]

        return None

    def _tag_resource(self, resource_arn: str, tags: list[dict[str, str]]) -> None:
        self.tagger.tag_resource(resource_arn, tags)

    def _list_tags_for_resource(self, resource_arn: str) -> list[dict[str, str]]:
        if self.tagger.has_tags(resource_arn):
            return self.tagger.list_tags_for_resource(resource_arn)["Tags"]
        return []


servicecatalog_backends = BackendDict(ServiceCatalogBackend, "servicecatalog")
